////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

#ifndef SFML_RENDERWINDOW_HPP
#define SFML_RENDERWINDOW_HPP

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Graphics/Color.hpp>
#include <SFML/Graphics/View.hpp>
#include <SFML/Graphics/Rect.hpp>
#include <SFML/Window/Window.hpp>
#include <string>

class sfDrawable;


////////////////////////////////////////////////////////////
/// Simple wrapper for sfWindow that allows easy
/// 2D rendering
////////////////////////////////////////////////////////////
class SFML_API sfRenderWindow : private sfWindow
{
public :

    using sfWindow::GetWidth;
    using sfWindow::GetHeight;
    using sfWindow::GetInput;
    using sfWindow::GetFrameTime;
    using sfWindow::UseVerticalSync;
    using sfWindow::ShowMouseCursor;

    ////////////////////////////////////////////////////////////
    /// Default constructor
    ///
    ////////////////////////////////////////////////////////////
    sfRenderWindow();

    ////////////////////////////////////////////////////////////
    /// Construct the window
    ///
    /// \param Mode :       Video mode to use
    /// \param Title :      Title of the window
    /// \param Fullscreen : True to set fullscreen, false to stay in windowed mode
    ///
    ////////////////////////////////////////////////////////////
    sfRenderWindow(sfVideoMode Mode, const std::string& Title, bool Fullscreen);

    ////////////////////////////////////////////////////////////
    /// Construct the window from an existing control
    ///
    /// \param Handle : Platform-specific handle of the control
    ///
    ////////////////////////////////////////////////////////////
    sfRenderWindow(sfWindowHandle Handle);

    ////////////////////////////////////////////////////////////
    /// Destructor
    ///
    ////////////////////////////////////////////////////////////
    virtual ~sfRenderWindow();

    ////////////////////////////////////////////////////////////
    /// Create the window
    ///
    /// \param Mode :       Video mode to use
    /// \param Title :      Title of the window
    /// \param Fullscreen : True to set fullscreen, false to stay in windowed mode
    ///
    ////////////////////////////////////////////////////////////
    void Create(sfVideoMode Mode, const std::string& Title, bool Fullscreen);

    ////////////////////////////////////////////////////////////
    /// Create the window from an existing control
    ///
    /// \param Handle : Platform-specific handle of the control
    ///
    ////////////////////////////////////////////////////////////
    void Create(sfWindowHandle Handle);

    ////////////////////////////////////////////////////////////
    /// Display window content on screen
    ///
    ////////////////////////////////////////////////////////////
    void Display();

    ////////////////////////////////////////////////////////////
    /// Draw something on the window
    ///
    /// \param Object : Object to draw
    ///
    ////////////////////////////////////////////////////////////
    void Draw(sfDrawable& Object);

    ////////////////////////////////////////////////////////////
    /// Save the content of the window to a file
    ///
    /// \param Filename : Path of the file to save (file is overwritten if it already exists)
    ///
    ////////////////////////////////////////////////////////////
    void Capture(const std::string& Filename) const;

    ////////////////////////////////////////////////////////////
    /// Change the background color of the window
    ///
    /// \param Color : New background color
    ///
    ////////////////////////////////////////////////////////////
    void SetBackgroundColor(const sfColor& Color);

    ////////////////////////////////////////////////////////////
    /// Get the event on top of events stack, if any
    ///
    /// \param Event : Event to fill, if any
    ///
    /// \return True if an event was returned, false if events stack was empty
    ///
    ////////////////////////////////////////////////////////////
    bool GetEvent(sfEvent& Event);

    ////////////////////////////////////////////////////////////
    /// Change the current active view
    ///
    /// \param View : Pointer to the new view (pass NULL to set the default view)
    ///
    ////////////////////////////////////////////////////////////
    void SetView(const sfView* View);

    ////////////////////////////////////////////////////////////
    /// Get the current view rectangle
    ///
    /// \return Current view rectangle, in global coordinates
    ///
    ////////////////////////////////////////////////////////////
    const sfFloatRect& GetViewRect() const;

    ////////////////////////////////////////////////////////////
    /// Start custom OpenGL rendering
    ///
    ////////////////////////////////////////////////////////////
    void BeginOpenGL();

    ////////////////////////////////////////////////////////////
    /// End custom OpenGL rendering
    ///
    ////////////////////////////////////////////////////////////
    void EndOpenGL();

private :

    ////////////////////////////////////////////////////////////
    /// Initialize internal window
    ///
    ////////////////////////////////////////////////////////////
    void Initialize();

    ////////////////////////////////////////////////////////////
    // Member data
    ////////////////////////////////////////////////////////////
    sfColor     myBackgroundColor; ///< Background color
    sfView      myDefaultView;     ///< Default view
    sfFloatRect myCurrentRect;     ///< Rectangle corresponding to the current view
    bool        myOpenGLMode;      ///< True when we are between BeginOpenGL() and EndOpenGL()
};


#endif // SFML_RENDERWINDOW_HPP
