////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Audio/AudioDevice.hpp>
#include <SFML/Audio/AudioResource.hpp>
#include <algorithm>
#include <iostream>


namespace sf_private
{
////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
sfAudioDevice::sfAudioDevice()
{
    // Create the device
    myDevice = alcOpenDevice(NULL);

    if (myDevice)
    {
        // Create the context
        myContext = alcCreateContext(myDevice, NULL);

        if (myContext)
        {
            // Set the context as the current one (we'll only need one)
            alcMakeContextCurrent(myContext);

            // Setup a default listener
            float ListenerPos[] = {0.f, 0.f, 0.f};
            float ListenerVel[] = {0.f, 0.f, 0.f};
            float ListenerOri[] = {0.f, 0.f, 0.f, 0.f, 1.f, 0.f};
            sfALCheck(alListenerfv(AL_POSITION,    ListenerPos));
            sfALCheck(alListenerfv(AL_VELOCITY,    ListenerVel));
            sfALCheck(alListenerfv(AL_ORIENTATION, ListenerOri));
        }
        else
        {
            std::cerr << "Failed to create audio context" << std::endl;
        }
    }
    else
    {
        std::cerr << "Failed to open audio device" << std::endl;
    }
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
sfAudioDevice::~sfAudioDevice()
{
    // Destroy all device-dependant resources
    std::for_each(myResources.begin(), myResources.end(), std::mem_fun(&sfAudioResource::DestroyAudioResources));

    // Destroy the context
    alcMakeContextCurrent(NULL);
    if (myContext)
        alcDestroyContext(myContext);

    // Destroy the device
    if (myDevice)
        alcCloseDevice(myDevice);
}


////////////////////////////////////////////////////////////
/// Get the unique instance of the class
////////////////////////////////////////////////////////////
sfAudioDevice& sfAudioDevice::GetInstance()
{
    static sfAudioDevice Instance;

    return Instance;
}


////////////////////////////////////////////////////////////
/// Add an audio resource
////////////////////////////////////////////////////////////
void sfAudioDevice::AddAudioResource(sfAudioResource* Resource)
{
    if (Resource)
        myResources.insert(Resource);
}


////////////////////////////////////////////////////////////
/// Remove an audio resource
////////////////////////////////////////////////////////////
void sfAudioDevice::RemoveAudioResource(sfAudioResource* Resource)
{
    if (Resource)
        myResources.erase(Resource);
}


////////////////////////////////////////////////////////////
/// Get the OpenAL audio device
////////////////////////////////////////////////////////////
ALCdevice* sfAudioDevice::GetDevice() const
{
    return myDevice;
}


////////////////////////////////////////////////////////////
/// Get the OpenAL format that matches the given number of channels
////////////////////////////////////////////////////////////
ALenum sfAudioDevice::GetFormatFromChannelsCount(unsigned int ChannelsCount) const
{
    // Find the good format according to the number of channels
    switch (ChannelsCount)
    {
        case 1 : return AL_FORMAT_MONO16;
        case 2 : return AL_FORMAT_STEREO16;
        case 4 : return alGetEnumValue("AL_FORMAT_QUAD16");
        case 6 : return alGetEnumValue("AL_FORMAT_51CHN16");
        case 7 : return alGetEnumValue("AL_FORMAT_61CHN16");
        case 8 : return alGetEnumValue("AL_FORMAT_71CHN16");
    }

    return 0;
}

} // namespace sf_private
