////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Audio/Sound.hpp>
#include <SFML/Audio/SoundBuffer.hpp>
#include <SFML/Audio/OpenAL.hpp>


////////////////////////////////////////////////////////////
/// Construct the sound from its parameters
////////////////////////////////////////////////////////////
sfSound::sfSound(const sfSoundBuffer* Buffer, bool Loop, float Pitch, float Volume, float X, float Y, float Z) :
myBuffer(Buffer)
{
    sfALCheck(alGenSources(1, &mySource));

    sfALCheck(alSourcei (mySource, AL_BUFFER,   myBuffer ? myBuffer->myBuffer : 0));
    sfALCheck(alSourcei (mySource, AL_LOOPING,  Loop));
    sfALCheck(alSourcef (mySource, AL_PITCH,    Pitch));
    sfALCheck(alSourcef (mySource, AL_GAIN,     Volume * 0.01f));
    sfALCheck(alSource3f(mySource, AL_POSITION, X, Y, Z));

    // Seems to fix the bug of mono sounds not being played due to spacialization,
    // although the listener and the sound are at the same position
    sfALCheck(alSourcef (mySource, AL_ROLLOFF_FACTOR, 0.99f));
}


////////////////////////////////////////////////////////////
/// Copy constructor
////////////////////////////////////////////////////////////
sfSound::sfSound(const sfSound& Copy) :
sf_private::sfAudioResource(Copy),
myBuffer(Copy.myBuffer)
{
    sfALCheck(alGenSources(1, &mySource));

    sfALCheck(alSourcei(mySource, AL_BUFFER,  myBuffer ? myBuffer->myBuffer : 0));
    sfALCheck(alSourcei(mySource, AL_LOOPING, Copy.GetLoop()));
    sfALCheck(alSourcef(mySource, AL_PITCH,   Copy.GetPitch()));
    sfALCheck(alSourcef(mySource, AL_GAIN,    Copy.GetVolume() * 0.01f));

    float X, Y, Z;
    Copy.GetPosition(X, Y, Z);
    sfALCheck(alSource3f(mySource, AL_POSITION, X, Y, Z));

    // Seems to fix the bug of mono sounds not being played due to spacialization,
    // although the listener and the sound are at the same position
    sfALCheck(alSourcef (mySource, AL_ROLLOFF_FACTOR, 0.99f));
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
sfSound::~sfSound()
{
    Stop();
    DestroyAudioResources();
}


////////////////////////////////////////////////////////////
/// Play the sound
////////////////////////////////////////////////////////////
void sfSound::Play()
{
    sfALCheck(alSourcePlay(mySource));
}


////////////////////////////////////////////////////////////
/// Pause the sound
////////////////////////////////////////////////////////////
void sfSound::Pause()
{
    sfALCheck(alSourcePause(mySource));
}


////////////////////////////////////////////////////////////
/// Stop the sound
////////////////////////////////////////////////////////////
void sfSound::Stop()
{
    sfALCheck(alSourceStop(mySource));
}


////////////////////////////////////////////////////////////
/// Set the source buffer
////////////////////////////////////////////////////////////
void sfSound::SetBuffer(const sfSoundBuffer* Buffer)
{
    myBuffer = Buffer;
    sfALCheck(alSourcei(mySource, AL_BUFFER, myBuffer ? myBuffer->myBuffer : 0));
}


////////////////////////////////////////////////////////////
/// Set the sound loop state
////////////////////////////////////////////////////////////
void sfSound::SetLoop(bool Loop)
{
    sfALCheck(alSourcei(mySource, AL_LOOPING, Loop));
}


////////////////////////////////////////////////////////////
/// Set the sound pitch
////////////////////////////////////////////////////////////
void sfSound::SetPitch(float Pitch)
{
    sfALCheck(alSourcef(mySource, AL_PITCH, Pitch));
}


////////////////////////////////////////////////////////////
/// Set the sound volume
////////////////////////////////////////////////////////////
void sfSound::SetVolume(float Volume)
{
    sfALCheck(alSourcef(mySource, AL_GAIN, Volume * 0.01f));
}


////////////////////////////////////////////////////////////
/// Set the sound position
////////////////////////////////////////////////////////////
void sfSound::SetPosition(float X, float Y, float Z)
{
    sfALCheck(alSource3f(mySource, AL_POSITION, X, Y, Z));
}


////////////////////////////////////////////////////////////
/// Get the source buffer
////////////////////////////////////////////////////////////
const sfSoundBuffer* sfSound::GetBuffer() const
{
    return myBuffer;
}


////////////////////////////////////////////////////////////
/// Tell whether or not the sound is looping
////////////////////////////////////////////////////////////
bool sfSound::GetLoop() const
{
    ALint Loop;
    sfALCheck(alGetSourcei(mySource, AL_LOOPING, &Loop));

    return Loop != 0;
}


////////////////////////////////////////////////////////////
/// Get the pitch
////////////////////////////////////////////////////////////
float sfSound::GetPitch() const
{
    ALfloat Pitch;
    sfALCheck(alGetSourcef(mySource, AL_PITCH, &Pitch));

    return Pitch;
}


////////////////////////////////////////////////////////////
/// Get the volume
////////////////////////////////////////////////////////////
float sfSound::GetVolume() const
{
    ALfloat Gain;
    sfALCheck(alGetSourcef(mySource, AL_GAIN, &Gain));

    return Gain * 100.f;
}


////////////////////////////////////////////////////////////
/// Get the sound position
////////////////////////////////////////////////////////////
void sfSound::GetPosition(float& X, float& Y, float& Z) const
{
    sfALCheck(alGetSource3f(mySource, AL_POSITION, &X, &Y, &Z));
}


////////////////////////////////////////////////////////////
/// Get the status of the sound (stopped, paused, playing)
////////////////////////////////////////////////////////////
sfSound::Status sfSound::GetStatus() const
{
    ALint State;
    sfALCheck(alGetSourcei(mySource, AL_SOURCE_STATE, &State));

    switch (State)
    {
        case AL_INITIAL :
        case AL_STOPPED : return Stopped;
        case AL_PAUSED :  return Paused;
        case AL_PLAYING : return Playing;
    }

    return Stopped;
}


////////////////////////////////////////////////////////////
/// Get the current playing position of the sound
////////////////////////////////////////////////////////////
float sfSound::GetPlayingOffset() const
{
    ALfloat Seconds = 0.f;
    sfALCheck(alGetSourcef(mySource, AL_SEC_OFFSET, &Seconds));

    return Seconds;
}


////////////////////////////////////////////////////////////
/// Assignment operator
////////////////////////////////////////////////////////////
sfSound& sfSound::operator =(const sfSound& Other)
{
    sfSound Temp(Other);

    std::swap(mySource, Temp.mySource);
    std::swap(myBuffer, Temp.myBuffer);

    return *this;
}


////////////////////////////////////////////////////////////
/// /see sfAudioResource::DestroyAudioResources
////////////////////////////////////////////////////////////
void sfSound::DestroyAudioResources()
{
    if (mySource)
    {
        sfALCheck(alSourcei(mySource, AL_BUFFER, 0));
        sfALCheck(alDeleteSources(1, &mySource));
        mySource = 0;
    }
}
