////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Audio/SoundBuffer.hpp>
#include <SFML/Audio/SoundFile.hpp>
#include <SFML/Audio/AudioDevice.hpp>
#include <SFML/Audio/OpenAL.hpp>
#include <iostream>
#include <memory>


////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
sfSoundBuffer::sfSoundBuffer() :
myBuffer  (0),
myDuration(0.f)
{
    // Create the buffer
    sfALCheck(alGenBuffers(1, &myBuffer));
}


////////////////////////////////////////////////////////////
/// Copy constructor
////////////////////////////////////////////////////////////
sfSoundBuffer::sfSoundBuffer(const sfSoundBuffer& Copy) :
sf_private::sfAudioResource(Copy),
myBuffer  (0),
mySamples (Copy.mySamples),
myDuration(Copy.myDuration)
{
    // Create the buffer
    sfALCheck(alGenBuffers(1, &myBuffer));

    // Update the internal buffer with the new samples
    Update(Copy.GetChannelsCount(), Copy.GetSampleRate());
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
sfSoundBuffer::~sfSoundBuffer()
{
    DestroyAudioResources();
}


////////////////////////////////////////////////////////////
/// Load the sound buffer from a file
////////////////////////////////////////////////////////////
bool sfSoundBuffer::LoadFromFile(const std::string& Filename)
{
    // Create the sound file
    std::auto_ptr<sf_private::sfSoundFile> File(sf_private::sfSoundFile::Create(Filename));

    // Open the sound file
    std::size_t  NbSamples;
    unsigned int ChannelsCount, SampleRate;
    if (File->OpenRead(Filename, NbSamples, ChannelsCount, SampleRate))
    {
        // Read the samples from the opened file
        mySamples.resize(NbSamples);
        if (File->Read(&mySamples[0], NbSamples) == NbSamples)
        {
            // Update the internal buffer with the new samples
            return Update(ChannelsCount, SampleRate);
        }
        else
        {
            // Error...
            std::cerr << "Failed to read audio data from file \"" << Filename << "\"" << std::endl;

            return false;
        }
    }
    else
    {
        // Error...
        std::cerr << "Failed to load sound buffer from file \"" << Filename << "\"" << std::endl;

        return false;
    }
}


////////////////////////////////////////////////////////////
/// Load the sound buffer from memory
////////////////////////////////////////////////////////////
bool sfSoundBuffer::LoadFromMemory(const sfInt16* Samples, std::size_t SamplesCount, unsigned int ChannelsCount, unsigned int SampleRate)
{
    if (Samples && SamplesCount && ChannelsCount && SampleRate)
    {
        // Copy the new audio samples
        mySamples.assign(Samples, Samples + SamplesCount);

        // Update the internal buffer with the new samples
        return Update(ChannelsCount, SampleRate);
    }
    else
    {
        // Error...
        std::cerr << "Failed to load sound buffer from memory ("
                  << "Samples : "        << Samples       << ", "
                  << "Samples count : "  << SamplesCount  << ", "
                  << "Channels count : " << ChannelsCount << ", "
                  << "Sample rate : "    << SampleRate    << ")"
                  << std::endl;

        return false;
    }
}


////////////////////////////////////////////////////////////
/// Save the sound buffer to a file
////////////////////////////////////////////////////////////
bool sfSoundBuffer::SaveToFile(const std::string& Filename) const
{
    // Create the sound file
    std::auto_ptr<sf_private::sfSoundFile> File(sf_private::sfSoundFile::Create(Filename));

    // Open a sound file
    if (File->OpenWrite(Filename, GetChannelsCount(), GetSampleRate()))
    {
        // Write the samples to the opened file
        File->Write(&mySamples[0], mySamples.size());

        return true;
    }
    else
    {
        // Error...
        std::cerr << "Failed to save sound buffer to file \"" << Filename << "\"" << std::endl;

        return false;
    }
}


////////////////////////////////////////////////////////////
/// Return the sound samples
////////////////////////////////////////////////////////////
const sfInt16* sfSoundBuffer::GetSamples() const
{
    return mySamples.empty() ? NULL : &mySamples[0];
}


////////////////////////////////////////////////////////////
/// Return the samples count
////////////////////////////////////////////////////////////
std::size_t sfSoundBuffer::GetSamplesCount() const
{
    return mySamples.size();
}


////////////////////////////////////////////////////////////
/// Get the sample rate
////////////////////////////////////////////////////////////
unsigned int sfSoundBuffer::GetSampleRate() const
{
    ALint SampleRate;
    sfALCheck(alGetBufferi(myBuffer, AL_FREQUENCY, &SampleRate));

    return SampleRate;
}


////////////////////////////////////////////////////////////
/// Return the number of channels (1 = mono, 2 = stereo, ...)
////////////////////////////////////////////////////////////
unsigned int sfSoundBuffer::GetChannelsCount() const
{
    ALint ChannelsCount;
    sfALCheck(alGetBufferi(myBuffer, AL_CHANNELS, &ChannelsCount));

    return ChannelsCount;
}


////////////////////////////////////////////////////////////
/// Get the sound duration
////////////////////////////////////////////////////////////
float sfSoundBuffer::GetDuration() const
{
    return myDuration;
}


////////////////////////////////////////////////////////////
/// Assignment operator
////////////////////////////////////////////////////////////
sfSoundBuffer& sfSoundBuffer::operator =(const sfSoundBuffer& Other)
{
    sfSoundBuffer Temp(Other);

    mySamples.swap(Temp.mySamples);
    std::swap(myBuffer,   Temp.myBuffer);
    std::swap(myDuration, Temp.myDuration);

    return *this;
}


////////////////////////////////////////////////////////////
/// Update the internal buffer with the audio samples
////////////////////////////////////////////////////////////
bool sfSoundBuffer::Update(unsigned int ChannelsCount, unsigned int SampleRate)
{
    // Check parameters
    if (!SampleRate || !ChannelsCount || mySamples.empty())
        return false;

    // Find the good format according to the number of channels
    ALenum Format = sf_private::sfAudioDevice::GetInstance().GetFormatFromChannelsCount(ChannelsCount);

    // Check if the format is valid
    if (Format == 0)
    {
        std::cerr << "Unsupported number of channels (" << ChannelsCount << ")" << std::endl;
        return false;
    }

    // Fill the buffer
    ALsizei Size = static_cast<ALsizei>(mySamples.size()) * sizeof(sfInt16);
    sfALCheck(alBufferData(myBuffer, Format, &mySamples[0], Size, SampleRate));

    // Compute the duration
    myDuration = static_cast<float>(mySamples.size()) / SampleRate / ChannelsCount;

    return true;
}


////////////////////////////////////////////////////////////
/// /see sfAudioResource::DestroyAudioResources
////////////////////////////////////////////////////////////
void sfSoundBuffer::DestroyAudioResources()
{
    if (myBuffer)
    {
        sfALCheck(alDeleteBuffers(1, &myBuffer));
        myBuffer = 0;
    }
}
