////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Graphics/Drawable.hpp>
#include <SFML/Graphics/OpenGL.hpp>


////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
sfDrawable::sfDrawable(float Left, float Top, float Scale, float Rotation, const sfColor& Color) :
myLeft    (Left),
myTop     (Top),
myScale   (Scale),
myRotation(Rotation),
myCenterX (0.f),
myCenterY (0.f),
myColor   (Color)
{

}


////////////////////////////////////////////////////////////
/// Virtual destructor
////////////////////////////////////////////////////////////
sfDrawable::~sfDrawable()
{
    // Nothing to do
}


////////////////////////////////////////////////////////////
/// Set the left position of the object
////////////////////////////////////////////////////////////
void sfDrawable::SetLeft(float Left)
{
    myLeft = Left;
}


////////////////////////////////////////////////////////////
/// Set the top position of the object
////////////////////////////////////////////////////////////
void sfDrawable::SetTop(float Top)
{
    myTop = Top;
}


////////////////////////////////////////////////////////////
/// Set the uniform scale of the object
////////////////////////////////////////////////////////////
void sfDrawable::SetScale(float Scale)
{
    if (Scale > 0.f)
        myScale = Scale;
}


////////////////////////////////////////////////////////////
/// Set the orientation of the object
////////////////////////////////////////////////////////////
void sfDrawable::SetRotation(float Rotation)
{
    myRotation = Rotation;
}


////////////////////////////////////////////////////////////
/// Set the center of rotation, in coordinates relative to the
/// object
////////////////////////////////////////////////////////////
void sfDrawable::SetRotationCenter(float X, float Y)
{
    myCenterX = X;
    myCenterY = Y;
}


////////////////////////////////////////////////////////////
/// Set the color of the object
////////////////////////////////////////////////////////////
void sfDrawable::SetColor(const sfColor& Color)
{
    myColor = Color;
}


////////////////////////////////////////////////////////////
/// Get the left position of the object
////////////////////////////////////////////////////////////
float sfDrawable::GetLeft() const
{
    return myLeft;
}


////////////////////////////////////////////////////////////
/// Get the top position of the object
////////////////////////////////////////////////////////////
float sfDrawable::GetTop() const
{
    return myTop;
}


////////////////////////////////////////////////////////////
/// Get the uniform scale of the object
////////////////////////////////////////////////////////////
float sfDrawable::GetScale() const
{
    return myScale;
}


////////////////////////////////////////////////////////////
/// Get the orientation of the object
////////////////////////////////////////////////////////////
float sfDrawable::GetRotation() const
{
    return myRotation;
}


////////////////////////////////////////////////////////////
/// Get the color of the object
////////////////////////////////////////////////////////////
const sfColor& sfDrawable::GetColor() const
{
    return myColor;
}


////////////////////////////////////////////////////////////
/// Move the object
////////////////////////////////////////////////////////////
void sfDrawable::Move(float OffsetX, float OffsetY)
{
    myLeft += OffsetX;
    myTop  += OffsetY;
}


////////////////////////////////////////////////////////////
/// Scale the object
////////////////////////////////////////////////////////////
void sfDrawable::Scale(float Factor)
{
    if (Factor > 0.f)
        myScale *= Factor;
}


////////////////////////////////////////////////////////////
/// Rotate the object
////////////////////////////////////////////////////////////
void sfDrawable::Rotate(float Angle)
{
    myRotation += Angle;
}


////////////////////////////////////////////////////////////
/// Draw the object into the specified window
////////////////////////////////////////////////////////////
void sfDrawable::Draw(sfRenderWindow& Window)
{
    // Reset transformations
    sfGLCheck(glMatrixMode(GL_MODELVIEW));
    sfGLCheck(glLoadIdentity());

    // Setup alpha-blending
    sfGLCheck(glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA));

    // Set color
    sfGLCheck(glColor4ub(myColor.r, myColor.g, myColor.b, myColor.a));

    // Set transformations
    sfGLCheck(glTranslatef(myLeft + myCenterX, myTop + myCenterY, 0));
    sfGLCheck(glRotatef(-myRotation, 0, 0, 1));
    sfGLCheck(glTranslatef(-myCenterX, -myCenterY, 0));
    sfGLCheck(glScalef(myScale, myScale, 1));

    // Let the derived class render the object geometry
    Render(Window);
}
