////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Graphics/FontManager.hpp>
#include <SFML/Graphics/Color.hpp>
#include <SFML/Graphics/GraphicsDevice.hpp>
#include <SFML/Graphics/Image.hpp>
#include <SFML/Window/OpenGLCaps.hpp>
#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H
#include <iostream>
#include <vector>


namespace
{
    ////////////////////////////////////////////////////////////
    /// The global instance of the FreeType library
    ////////////////////////////////////////////////////////////
    FT_Library Library = NULL;

    ////////////////////////////////////////////////////////////
    /// Get a description from a FT error code
    ////////////////////////////////////////////////////////////
    std::string GetErrorDesc(FT_Error Error)
    {
        switch (Error)
        {
            // Generic errors
            case FT_Err_Cannot_Open_Resource :      return "cannot open resource";
            case FT_Err_Unknown_File_Format :       return "unknown file format";
            case FT_Err_Invalid_File_Format :       return "broken file";
            case FT_Err_Invalid_Version :           return "invalid FreeType version";
            case FT_Err_Lower_Module_Version :      return "module version is too low";
            case FT_Err_Invalid_Argument :          return "invalid argument";
            case FT_Err_Unimplemented_Feature :     return "unimplemented feature";
            case FT_Err_Invalid_Table :             return "broken table";
            case FT_Err_Invalid_Offset :            return "broken offset within table";

            // Glyph / character errors
            case FT_Err_Invalid_Glyph_Index :       return "invalid glyph index";
            case FT_Err_Invalid_Character_Code :    return "invalid character code";
            case FT_Err_Invalid_Glyph_Format :      return "unsupported glyph image format";
            case FT_Err_Cannot_Render_Glyph :       return "cannot render this glyph format";
            case FT_Err_Invalid_Outline :           return "invalid outline";
            case FT_Err_Invalid_Composite :         return "invalid composite glyph";
            case FT_Err_Too_Many_Hints :            return "too many hints";
            case FT_Err_Invalid_Pixel_Size :        return "invalid pixel size";

            // Handle errors
            case FT_Err_Invalid_Handle :            return "invalid object handle";
            case FT_Err_Invalid_Library_Handle :    return "invalid library handle";
            case FT_Err_Invalid_Driver_Handle :     return "invalid module handle";
            case FT_Err_Invalid_Face_Handle :       return "invalid face handle";
            case FT_Err_Invalid_Size_Handle :       return "invalid size handle";
            case FT_Err_Invalid_Slot_Handle :       return "invalid glyph slot handle";
            case FT_Err_Invalid_CharMap_Handle :    return "invalid charmap handle";
            case FT_Err_Invalid_Cache_Handle :      return "invalid cache manager handle";
            case FT_Err_Invalid_Stream_Handle :     return "invalid stream handle";

            // Driver errors
            case FT_Err_Too_Many_Drivers :          return "too many modules";
            case FT_Err_Too_Many_Extensions :       return "too many extensions";

            // Memory errors
            case FT_Err_Out_Of_Memory :             return "out of memory";
            case FT_Err_Unlisted_Object :           return "unlisted object";

            // Stream errors
            case FT_Err_Cannot_Open_Stream :        return "cannot open stream";
            case FT_Err_Invalid_Stream_Seek :       return "invalid stream seek";
            case FT_Err_Invalid_Stream_Skip :       return "invalid stream skip";
            case FT_Err_Invalid_Stream_Read :       return "invalid stream read";
            case FT_Err_Invalid_Stream_Operation :  return "invalid stream operation";
            case FT_Err_Invalid_Frame_Operation :   return "invalid frame operation";
            case FT_Err_Nested_Frame_Access :       return "nested frame access";
            case FT_Err_Invalid_Frame_Read :        return "invalid frame read";

            // Raster errors
            case FT_Err_Raster_Uninitialized :      return "raster uninitialized";
            case FT_Err_Raster_Corrupted :          return "raster corrupted";
            case FT_Err_Raster_Overflow :           return "raster overflow";
            case FT_Err_Raster_Negative_Height :    return "negative height while rastering";

            // Cache errors
            case FT_Err_Too_Many_Caches :           return "too many registered caches";

            // TrueType and SFNT errors
            case FT_Err_Invalid_Opcode :            return "invalid opcode";
            case FT_Err_Too_Few_Arguments :         return "too few arguments";
            case FT_Err_Stack_Overflow :            return "stack overflow";
            case FT_Err_Code_Overflow :             return "code overflow";
            case FT_Err_Bad_Argument :              return "bad argument";
            case FT_Err_Divide_By_Zero :            return "division by zero";
            case FT_Err_Invalid_Reference :         return "invalid reference";
            case FT_Err_Debug_OpCode :              return "found debug opcode";
            case FT_Err_ENDF_In_Exec_Stream :       return "found ENDF opcode in execution stream";
            case FT_Err_Nested_DEFS :               return "nested DEFS";
            case FT_Err_Invalid_CodeRange :         return "invalid code range";
            case FT_Err_Execution_Too_Long :        return "execution context too long";
            case FT_Err_Too_Many_Function_Defs :    return "too many function definitions";
            case FT_Err_Too_Many_Instruction_Defs : return "too many instruction definitions";
            case FT_Err_Table_Missing :             return "SFNT font table missing";
            case FT_Err_Horiz_Header_Missing :      return "horizontal header (hhea) table missing";
            case FT_Err_Locations_Missing :         return "locations (loca) table missing";
            case FT_Err_Name_Table_Missing :        return "name table missing";
            case FT_Err_CMap_Table_Missing :        return "character map (cmap) table missing";
            case FT_Err_Hmtx_Table_Missing :        return "horizontal metrics (hmtx) table missing";
            case FT_Err_Post_Table_Missing :        return "PostScript (post) table missing";
            case FT_Err_Invalid_Horiz_Metrics :     return "invalid horizontal metrics";
            case FT_Err_Invalid_CharMap_Format :    return "invalid character map (cmap) format";
            case FT_Err_Invalid_PPem :              return "invalid ppem value";
            case FT_Err_Invalid_Vert_Metrics :      return "invalid vertical metrics";
            case FT_Err_Could_Not_Find_Context :    return "could not find context";
            case FT_Err_Invalid_Post_Table_Format : return "invalid PostScript (post) table format";
            case FT_Err_Invalid_Post_Table :        return "invalid PostScript (post) table";

            // CCF, CID and Type 1 errors
            case FT_Err_Syntax_Error :              return "opcode syntax error";
            case FT_Err_Stack_Underflow :           return "argument stack underflow";
            case FT_Err_Ignore :                    return "ignore";

            // BDF errors
            case FT_Err_Missing_Startfont_Field :   return "`STARTFONT' field missing";
            case FT_Err_Missing_Font_Field :        return "`FONT' field missing";
            case FT_Err_Missing_Size_Field :        return "`SIZE' field missing";
            case FT_Err_Missing_Chars_Field :       return "`CHARS' field missing";
            case FT_Err_Missing_Startchar_Field :   return "`STARTCHAR' field missing";
            case FT_Err_Missing_Encoding_Field :    return "`ENCODING' field missing";
            case FT_Err_Missing_Bbx_Field :         return "`BBX' field missing";
        }

        return "";
    }
}

namespace sf_private
{
// Default font bitmap and description are stored in a header file as a sequence of arrays
#include <SFML/Graphics/DefaultFont.hpp>


////////////////////////////////////////////////////////////
/// Get the unique instance of the class
////////////////////////////////////////////////////////////
sfFontManager& sfFontManager::GetInstance()
{
    static sfFontManager Instance;

    return Instance;
}


////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
sfFontManager::sfFontManager()
{
    // Initialize FreeType library
    FT_Error Error = FT_Init_FreeType(&Library);
    if (Error)
    {
        std::cerr << "Failed to initialize FreeType library (error code : " << Error << ")" << std::endl;
        return;
    }

    // Create the default font
    CreateDefaultFont();
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
sfFontManager::~sfFontManager()
{
    // Shutdown FreeType library
    if (Library)
        FT_Done_FreeType(Library);
}


////////////////////////////////////////////////////////////
/// Get a bitmap font from a font file (create it if doesn't exist)
////////////////////////////////////////////////////////////
const sfFontManager::Font& sfFontManager::GetBitmapFont(const std::string& Filename, unsigned int CharSize)
{
    // Check if freetype library is correctly initialized
    if (!Library)
        return myFonts["default"];

    // If font is "default" or empty string, return default font
    if ((Filename == "") || (Filename == "default"))
        return myFonts["default"];

    // If font is already loaded and char size is big enough, just return it
    FontTable::iterator It = myFonts.find(Filename);
    if ((It != myFonts.end()) && (It->second.CharSize >= CharSize))
        return It->second;

    // Clamp CharSize to make sure we won't have textures too big
    int MaxSize = sfOpenGLCaps::GetMaxTextureSize();
    if ((int)CharSize >= MaxSize / 8) CharSize = MaxSize / 8;

    unsigned int Left      = 0;
    unsigned int Top       = 0;
    unsigned int TexWidth  = sfImage::GetValidTextureSize(CharSize * 8);
    unsigned int TexHeight = sfImage::GetValidTextureSize(CharSize * 8);
    std::vector<unsigned int> Tops(TexWidth, 0);

    // If font name has not been found, we create a new font description
    Font CurFont;
    CurFont.CharSize = CharSize;
    CurFont.Image.Create(TexWidth, TexHeight, sfColor(0, 0, 0, 0));

    // Create a new font face from specified file
    FT_Face FontFace;
    FT_Error Error = FT_New_Face(Library, Filename.c_str(), 0, &FontFace);
    if (Error)
    {
        std::cerr << "Error loading font \"" << Filename << "\" (" << GetErrorDesc(Error) << ")" << std::endl;
        return myFonts["default"];
    }

    // Setup font size
    Error = FT_Set_Pixel_Sizes(FontFace, CharSize, CharSize);
    if (Error)
    {
        std::cerr << "Error loading font \"" << Filename << "\" (" << GetErrorDesc(Error) << ")" << std::endl;
        return myFonts["default"];
    }

    // Render characters set to a bitmap (printable characters start at code 31)
    sfIntRect Coords[256];
    for (int i = 31; i < 256; ++i)
    {
        // Load the glyph corresponding to the current character
        Error = FT_Load_Char(FontFace, i, FT_LOAD_DEFAULT);
        if (Error)
        {
            std::cerr << "Error loading font \"" << Filename << "\" (" << GetErrorDesc(Error) << ")" << std::endl;
            return myFonts["default"];
        }

        // Convert the glyph to a bitmap
        FT_Glyph Glyph;
        Error = FT_Get_Glyph(FontFace->glyph, &Glyph);
        if (Error)
        {
            std::cerr << "Error loading font \"" << Filename << "\" (" << GetErrorDesc(Error) << ")" << std::endl;
            return myFonts["default"];
        }
        FT_Glyph_To_Bitmap(&Glyph, ft_render_mode_normal, 0, 1);
        FT_BitmapGlyph BitmapGlyph = (FT_BitmapGlyph)Glyph;
        FT_Bitmap& Bitmap = BitmapGlyph->bitmap;

        // TODO : handle other pixel modes
        if (Bitmap.pixel_mode != FT_PIXEL_MODE_GRAY)
        {
            std::cerr << "Error loading font \"" << Filename << "\" (pixel format not supported)" << std::endl;
            return myFonts["default"];
        }

        // Make sure we don't go over the texture width
        if (Left + Bitmap.width >= TexWidth)
            Left = 0;

        // Compute the top coordinate
        Top = Tops[Left];
        for (int x = 0; x < Bitmap.width; ++x)
            Top = std::max(Top, Tops[Left + x]);

        // Make sure we don't go over the texture height
        if (Top + Bitmap.rows >= TexHeight)
        {
            TexHeight *= 2;
            CurFont.Image.Resize(TexWidth, TexHeight, sfColor(0, 0, 0, 0));
        }

        // Store the character's position and size
        sfIntRect& Rect = CurFont.Rect[i];
        Rect.Left   = BitmapGlyph->left;
        Rect.Top    = -BitmapGlyph->top;
        Rect.Right  = Rect.Left + Bitmap.width;
        Rect.Bottom = Bitmap.rows - BitmapGlyph->top;
        CurFont.Advance[i] = FontFace->glyph->advance.x / 64;

        // Texture size may change, so let the texture coordinates be calculated later
        Coords[i] = sfIntRect(Left, Top, Left + Bitmap.width, Top + Bitmap.rows);

        // Draw the glyph into our bitmap font
        const sfUint8* Pixels = Bitmap.buffer;
        for (int y = 0; y < Bitmap.rows; ++y)
        {
            for (int x = 0; x < Bitmap.width; ++x)
            {
                CurFont.Image.SetPixel(x + Left, y + Top, sfColor(Pixels[x], Pixels[x], Pixels[x], Pixels[x]));
            }
            Pixels += Bitmap.pitch;
        }

        for (int x = 0; x < Bitmap.width; ++x)
            Tops[Left + x] = Top + Bitmap.rows;

        Left += Bitmap.width;
    }

    // Now that the texture has its final size, we can precompute texture coordinates
    for (int i = 31; i < 256; ++i)
        CurFont.Coord[i] = CurFont.Image.GetTexCoords(Coords[i]);

    // Update image after modifications
    CurFont.Image.Update();

    // Insert new font
    myFonts[Filename] = CurFont;

    return myFonts[Filename];
}


////////////////////////////////////////////////////////////
/// Create the default font
////////////////////////////////////////////////////////////
void sfFontManager::CreateDefaultFont()
{
    Font DefaultFont;

    // Load bitmap
    DefaultFont.Image.Create(256, 512);
    for (int j = 0; j < 512; ++j)
        for (int i = 0; i < 256; ++i)
        {
            sfUint8 Lum = DefaultFontBitmap[i + j * 256];
            DefaultFont.Image.SetPixel(i, j, sfColor(Lum, Lum, Lum, Lum));
        }
    DefaultFont.Image.Update();

    // Load positions
    for (int i = 0; i < 256; i++)
    {
        DefaultFont.Rect[i] = sfIntRect(DefaultFontRect[i * 4 + 0],
                                        DefaultFontRect[i * 4 + 1],
                                        DefaultFontRect[i * 4 + 2],
                                        DefaultFontRect[i * 4 + 3]);
    }

    // Load texture coordinates
    for (int i = 0; i < 256; i++)
    {
        DefaultFont.Coord[i] = sfFloatRect(DefaultFontCoord[i * 4 + 0],
                                           DefaultFontCoord[i * 4 + 1],
                                           DefaultFontCoord[i * 4 + 2],
                                           DefaultFontCoord[i * 4 + 3]);
    }

    // Load advances
    for (int i = 0; i < 256; i++)
        DefaultFont.Advance[i] = DefaultFontAdvance[i];

    // Load character size
    DefaultFont.CharSize = DefaultFontCharSize;

    // Insert font in font table
    myFonts.insert(FontTable::value_type("default", DefaultFont));
}

} // namespace sf_private
