////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Graphics/RenderWindow.hpp>
#include <SFML/Graphics/Drawable.hpp>
#include <SFML/Graphics/GraphicsDevice.hpp>
#include <SFML/Graphics/Image.hpp>
#include <SFML/Graphics/OpenGL.hpp>
#include <iostream>


////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
sfRenderWindow::sfRenderWindow() :
myBackgroundColor(sfColor(0, 0, 0, 255)),
myOpenGLMode     (false)
{

}


////////////////////////////////////////////////////////////
/// Construct the window
////////////////////////////////////////////////////////////
sfRenderWindow::sfRenderWindow(sfVideoMode Mode, const std::string& Title, bool Fullscreen) :
myBackgroundColor(sfColor(0, 0, 0, 255)),
myOpenGLMode     (false)
{
    Create(Mode, Title, Fullscreen);
}


////////////////////////////////////////////////////////////
/// Construct the window from an existing control
////////////////////////////////////////////////////////////
sfRenderWindow::sfRenderWindow(sfWindowHandle Handle) :
myBackgroundColor(sfColor(0, 0, 0, 255)),
myOpenGLMode     (false)
{
    Create(Handle);
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
sfRenderWindow::~sfRenderWindow()
{
    // Nothing to do...
}


////////////////////////////////////////////////////////////
/// Create the window
////////////////////////////////////////////////////////////
void sfRenderWindow::Create(sfVideoMode Mode, const std::string& Title, bool Fullscreen)
{
    // Let the base class do the job
    sfWindow::Create(Mode, Title, Fullscreen);

    // Initialize the rendering states
    Initialize();
}


////////////////////////////////////////////////////////////
/// Create the window from an existing control
////////////////////////////////////////////////////////////
void sfRenderWindow::Create(sfWindowHandle Handle)
{
    // Let base class do the job
    sfWindow::Create(Handle);

    // Initialize the rendering states
    Initialize();
}


////////////////////////////////////////////////////////////
/// Display window content on screen
////////////////////////////////////////////////////////////
void sfRenderWindow::Display()
{
    // Discard display if we are in OpenGL custom rendering mode (it would mess up the render states)
    if (myOpenGLMode)
    {
        std::cerr << "You cannot display a window while in OpenGL mode ; call EndOpenGL() before calling Display()" << std::endl;
        return;
    }

    // Set our window as the current target for rendering
    if (sfWindow::SetCurrent())
    {
        // Display backbuffer on screen
        sfWindow::Display();

        // Find which buffers we must clear
        GLbitfield                ClearBits  = GL_COLOR_BUFFER_BIT;
        if (GetDepthBits()   > 0) ClearBits |= GL_DEPTH_BUFFER_BIT;
        if (GetStencilBits() > 0) ClearBits |= GL_STENCIL_BUFFER_BIT;
        
        // Clear the color/depth/stencil buffers for next frame
        sfGLCheck(glClearColor(myBackgroundColor.r / 255.f,
                               myBackgroundColor.g / 255.f,
                               myBackgroundColor.b / 255.f,
                               myBackgroundColor.a / 255.f));
        sfGLCheck(glClear(ClearBits));
    }
}


////////////////////////////////////////////////////////////
/// Draw something on the window
////////////////////////////////////////////////////////////
void sfRenderWindow::Draw(sfDrawable& Object)
{
    // Discard draw if we are in OpenGL custom rendering mode (it would mess up the render states)
    if (myOpenGLMode)
    {
        std::cerr << "You cannot use SFML rendering while in OpenGL mode ; call EndOpenGL() before calling Draw()" << std::endl;
        return;
    }

    // Set our window as the current target for rendering
    if (sfWindow::SetCurrent())
    {
        // Let the object draw itself
        Object.Draw(*this);
    }
}


////////////////////////////////////////////////////////////
/// Save the content of the window to a file
////////////////////////////////////////////////////////////
void sfRenderWindow::Capture(const std::string& Filename) const
{
    // Set our window as the current target for rendering
    if (sfWindow::SetCurrent())
    {
        // Get the window dimensions
        const unsigned int Width  = GetWidth();
        const unsigned int Height = GetHeight();

        // Get pixels from the backbuffer
        std::vector<sfUint32> Pixels(Width * Height);
        glReadPixels(0, 0, Width, Height, GL_RGBA, GL_UNSIGNED_BYTE, &Pixels[0]);

        // Create an image from it and save it to disk
        sfImage Screen(Width, Height, &Pixels[0]);
        Screen.SaveToFile(Filename);
    }
}


////////////////////////////////////////////////////////////
/// Change the background color of the window
////////////////////////////////////////////////////////////
void sfRenderWindow::SetBackgroundColor(const sfColor& Color)
{
    myBackgroundColor = Color;
}


////////////////////////////////////////////////////////////
/// Get the event on top of events stack, if any
////////////////////////////////////////////////////////////
bool sfRenderWindow::GetEvent(sfEvent& Event)
{
    // Forward to base class
    if (sfWindow::GetEvent(Event))
    {
        // Intercept resize events, because we have to adjust the viewport to new size
        if (Event.Type == sfEvent::Resize)
            sfGLCheck(glViewport(0, 0, Event.Size.Width, Event.Size.Height));

        return true;
    }

    return false;
}


////////////////////////////////////////////////////////////
/// Change the current active view
////////////////////////////////////////////////////////////
void sfRenderWindow::SetView(const sfView* View)
{
    // Discard display if we are in OpenGL custom rendering mode (it would mess up the render states)
    if (myOpenGLMode)
    {
        std::cerr << "You cannot change the view while in OpenGL mode ; call EndOpenGL() before setting a new SFML view" << std::endl;
        return;
    }

    // Set our window as the current target for rendering
    if (sfWindow::SetCurrent())
    {
        // If a null pointer is passed, switch to default view
        if (View == NULL)
            View = &myDefaultView;

        // Compute view rectangle coordinates
        float X = View->Left + View->Width  / 2;
        float Y = View->Top  + View->Height / 2;
        float HalfWidth  = View->Width  / (2 * View->Zoom);
        float HalfHeight = View->Height / (2 * View->Zoom);

        // Store view rectangle for optimization purpose
        myCurrentRect = sfFloatRect(X - HalfWidth, Y - HalfHeight, X + HalfWidth, Y + HalfHeight);

        // Update the projection matrix according to the new view
        sfGLCheck(glMatrixMode(GL_PROJECTION));
        sfGLCheck(glLoadIdentity());
        sfGLCheck(glOrtho(myCurrentRect.Left, myCurrentRect.Right, myCurrentRect.Bottom, myCurrentRect.Top, -1, 1));
    }
}


////////////////////////////////////////////////////////////
/// Get current view rectangle
////////////////////////////////////////////////////////////
const sfFloatRect& sfRenderWindow::GetViewRect() const
{
    return myCurrentRect;
}


////////////////////////////////////////////////////////////
/// Start custom OpenGL rendering
////////////////////////////////////////////////////////////
void sfRenderWindow::BeginOpenGL()
{
    if (!myOpenGLMode && sfWindow::SetCurrent())
    {
        // Save current projection matrix
        sfGLCheck(glMatrixMode(GL_PROJECTION));
        sfGLCheck(glPushMatrix());

        // Save current render states
        sfGLCheck(glPushAttrib(GL_ALL_ATTRIB_BITS));

        // Set OpenGL mode flag
        myOpenGLMode = true;
    }
}


////////////////////////////////////////////////////////////
/// End custom OpenGL rendering
////////////////////////////////////////////////////////////
void sfRenderWindow::EndOpenGL()
{
    if (myOpenGLMode && sfWindow::SetCurrent())
    {
        // Restore render states
        sfGLCheck(glPopAttrib());

        // Restore projection matrix
        sfGLCheck(glMatrixMode(GL_PROJECTION));
        sfGLCheck(glPopMatrix());

        // Clear OpenGL mode flag
        myOpenGLMode = false;
    }
}


////////////////////////////////////////////////////////////
/// Initialize internal window
////////////////////////////////////////////////////////////
void sfRenderWindow::Initialize()
{
    if (sfWindow::SetCurrent())
    {
        // Set default OpenGL states
        sfGLCheck(glEnable(GL_ALPHA_TEST));
        sfGLCheck(glAlphaFunc(GL_GREATER, 0));
        sfGLCheck(glEnable(GL_BLEND));
        sfGLCheck(glEnable(GL_TEXTURE_2D));
        sfGLCheck(glDisable(GL_LIGHTING));
        sfGLCheck(glDisable(GL_DEPTH_TEST));
        sfGLCheck(glDisable(GL_CULL_FACE));
        sfGLCheck(glShadeModel(GL_SMOOTH));
        sfGLCheck(glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST));
        sfGLCheck(glClearDepth(1.f));
        sfGLCheck(glClearStencil(0));

        // Setup the default view
        myDefaultView.Left   = 0;
        myDefaultView.Top    = 0;
        myDefaultView.Width  = static_cast<float>(GetWidth());
        myDefaultView.Height = static_cast<float>(GetHeight());
        myDefaultView.Zoom   = 1.f;
        SetView(NULL);
    }
}
