////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Graphics/Sprite.hpp>
#include <SFML/Graphics/Image.hpp>
#include <SFML/Graphics/OpenGL.hpp>


////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
sfSprite::sfSprite() :
myImage  (NULL),
mySubRect(0, 0, 0, 0)
{

}


////////////////////////////////////////////////////////////
/// Construct the sprite from a source image
////////////////////////////////////////////////////////////
sfSprite::sfSprite(const sfImage& Image, float Left, float Top, float Scale, float Rotation, const sfColor& Color) :
sfDrawable(Left, Top, Scale, Rotation, Color),
myImage   (&Image),
mySubRect (0, 0, Image.GetWidth(), Image.GetHeight())
{

}


////////////////////////////////////////////////////////////
/// Set the image of the sprite
////////////////////////////////////////////////////////////
void sfSprite::SetImage(const sfImage& Image)
{
    // Assign the new image
    myImage = &Image;

    // If the sub-rectangle was null, we define it as the whole image
    if ((mySubRect.GetWidth() == 0) || (mySubRect.GetHeight() == 0))
        SetSubRect(sfIntRect(0, 0, Image.GetWidth(), Image.GetHeight()));
}


////////////////////////////////////////////////////////////
/// Set the sub-rectangle of the sprite inside the source image
////////////////////////////////////////////////////////////
void sfSprite::SetSubRect(const sfIntRect& SubRect)
{
    mySubRect = SubRect;
}


////////////////////////////////////////////////////////////
/// Get the source image of the sprite
////////////////////////////////////////////////////////////
const sfImage* sfSprite::GetImage() const
{
    return myImage;
}


////////////////////////////////////////////////////////////
/// Get the sub-rectangle of the sprite inside the source image
////////////////////////////////////////////////////////////
const sfIntRect& sfSprite::GetSubRect() const
{
    return mySubRect;
}


////////////////////////////////////////////////////////////
/// Get the sprite width
////////////////////////////////////////////////////////////
float sfSprite::GetWidth() const
{
    return mySubRect.GetWidth() * GetScale();
}


////////////////////////////////////////////////////////////
/// Get the sprite height
////////////////////////////////////////////////////////////
float sfSprite::GetHeight() const
{
    return mySubRect.GetHeight() * GetScale();
}


////////////////////////////////////////////////////////////
/// Get the color of a given pixel in the sprite
////////////////////////////////////////////////////////////
sfColor sfSprite::GetPixel(unsigned int X, unsigned int Y) const
{
    return myImage ? myImage->GetPixel(X + mySubRect.Left, Y + mySubRect.Top) : sfColor(0, 0, 0, 0);
}


////////////////////////////////////////////////////////////
/// /see sfDrawable::Render
////////////////////////////////////////////////////////////
void sfSprite::Render(sfRenderWindow&)
{
    // Check if the image is valid
    if (myImage)
    {
        // Get the sprite size
        float Width  = static_cast<float>(mySubRect.GetWidth());
        float Height = static_cast<float>(mySubRect.GetHeight());

        // Calculate the texture coordinates
        sfFloatRect TexCoords = myImage->GetTexCoords(mySubRect);

        // Set the texture
        myImage->Bind();

        // Draw the sprite
        glBegin(GL_QUADS);
            glTexCoord2f(TexCoords.Left,  TexCoords.Top);    glVertex2f(0,     0);
            glTexCoord2f(TexCoords.Left,  TexCoords.Bottom); glVertex2f(0,     Height);
            glTexCoord2f(TexCoords.Right, TexCoords.Bottom); glVertex2f(Width, Height);
            glTexCoord2f(TexCoords.Right, TexCoords.Top);    glVertex2f(Width, 0);
        glEnd();
    }
}
