////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Graphics/String.hpp>
#include <SFML/Graphics/FontManager.hpp>
#include <SFML/Graphics/Image.hpp>
#include <SFML/Graphics/OpenGL.hpp>


////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
sfString::sfString(const std::string& Text, const std::string& Font, float Size) :
myText(Text),
myFont(Font),
mySize(Size)
{

}


////////////////////////////////////////////////////////////
/// Preload a bitmap font (otherwise, it is done the first time the font is drawn)
////////////////////////////////////////////////////////////
void sfString::PreloadFont(const std::string& Font, float Char)
{
    // Requesting the font will make the font manager create it
    sf_private::sfFontManager::GetInstance().GetBitmapFont(Font, static_cast<unsigned int>(Char));
}


////////////////////////////////////////////////////////////
/// Set the text
////////////////////////////////////////////////////////////
void sfString::SetText(const std::string& Text)
{
    myText = Text;
}


////////////////////////////////////////////////////////////
/// Set the font of the string
////////////////////////////////////////////////////////////
void sfString::SetFont(const std::string& Font)
{
    myFont = Font;
}


////////////////////////////////////////////////////////////
/// Set the size of the string
////////////////////////////////////////////////////////////
void sfString::SetSize(float Size)
{
    mySize = Size;
}


////////////////////////////////////////////////////////////
/// Get the text
////////////////////////////////////////////////////////////
const std::string& sfString::GetText() const
{
    return myText;
}


////////////////////////////////////////////////////////////
/// Get the font used by the string
////////////////////////////////////////////////////////////
const std::string& sfString::GetFont() const
{
    return myFont;
}


////////////////////////////////////////////////////////////
/// Get the size of the characters
////////////////////////////////////////////////////////////
float sfString::GetSize() const
{
    return mySize;
}


////////////////////////////////////////////////////////////
/// Get the string rectangle on screen
////////////////////////////////////////////////////////////
sfFloatRect sfString::GetRect() const
{
    // Get the bitmap font from the font manager
    const sf_private::sfFontManager::Font& BitmapFont = sf_private::sfFontManager::GetInstance().GetBitmapFont(myFont, static_cast<unsigned int>(mySize));

    // Initial values
    float CurWidth = 0;
    float Width    = 0;
    float Height   = mySize;

    // Go through each character
    for (std::size_t i = 0; i < myText.size(); ++i)
    {
        // Get the current character
        unsigned char c = static_cast<unsigned char>(myText[i]);

        // Get the dimensions of the current character from the font description
        const float Advance = BitmapFont.Advance[c] * mySize / BitmapFont.CharSize;

        // Handle special characters...
        switch (c)
        {
            case ' ' :  CurWidth += Advance;    continue;
            case '\t' : CurWidth += mySize * 4; continue;
            case '\v' : Height   += mySize * 4; continue;

            case '\n' :
                Height += mySize;
                if (CurWidth > Width) Width = CurWidth;
                CurWidth = 0;
                continue;
        }

        // Advance to the next character
        CurWidth += Advance;
    }

    // In case there was no '\n' in the string
    if (Width < 1)
        Width = CurWidth;

    return sfFloatRect(GetLeft(), GetTop(), GetLeft() + Width, GetTop() + Height);
}


////////////////////////////////////////////////////////////
/// /see sfDrawable::Render
////////////////////////////////////////////////////////////
void sfString::Render(sfRenderWindow&)
{
    // No text, no rendering :)
    if (myText.empty())
        return;

    // Get the bitmap font from the font manager
    const sf_private::sfFontManager::Font& BitmapFont = sf_private::sfFontManager::GetInstance().GetBitmapFont(myFont, static_cast<unsigned int>(mySize));

    // Set the scaling factor to get the actual size
    float Factor = GetScale() * mySize / BitmapFont.CharSize;
    sfGLCheck(glScalef(Factor, Factor, 1.f));

    // Bind the font texture
    BitmapFont.Image.Bind();

    float X = 0;
    float Y = 0;

    // Draw the sprite
    glBegin(GL_QUADS);
    for (std::size_t i = 0; i < myText.size(); ++i)
    {
        // Get the current character
        unsigned char c = static_cast<unsigned char>(myText[i]);

        // Get the dimensions of the current character from font description
        const sfIntRect&   Rect     = BitmapFont.Rect[c];
        const sfFloatRect& Coord    = BitmapFont.Coord[c];
        const unsigned int AdvanceX = BitmapFont.Advance[c];
        const unsigned int AdvanceY = BitmapFont.CharSize;

        // Handle special characters
        switch (c)
        {
            case ' ' :  X += AdvanceX;        continue;
            case '\n' : Y += AdvanceY; X = 0; continue;
            case '\t' : X += AdvanceX * 4;    continue;
            case '\v' : Y += AdvanceY * 4;    continue;
        }

        // Draw a textured quad for the current character
        glTexCoord2f(Coord.Left,  Coord.Top);    glVertex2f(X + Rect.Left,  Y + Rect.Top    + AdvanceY);
        glTexCoord2f(Coord.Left,  Coord.Bottom); glVertex2f(X + Rect.Left,  Y + Rect.Bottom + AdvanceY);
        glTexCoord2f(Coord.Right, Coord.Bottom); glVertex2f(X + Rect.Right, Y + Rect.Bottom + AdvanceY);
        glTexCoord2f(Coord.Right, Coord.Top);    glVertex2f(X + Rect.Right, Y + Rect.Top    + AdvanceY);

        // Advance to the next character
        X += AdvanceX;
    }
    glEnd();
}
