////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Window/Input.hpp>


////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
sfInput::sfInput() :
myMouseX(0),
myMouseY(0)
{
    for (int i = 0; i < sfKey::Count; ++i)
        myKeys[i] = false;

    for (int i = 0; i < 3; ++i)
        myMouseButtons[i] = false;

    myJoystickX[0] = 0;
    myJoystickX[1] = 0;
    myJoystickY[0] = 0;
    myJoystickY[1] = 0;
    myJoystickZ[0] = 0;
    myJoystickZ[1] = 0;

    for (int i = 0; i < 8; ++i)
    {
        myJoystickButtons[0][i] = false;
        myJoystickButtons[1][i] = false;
    }
}


////////////////////////////////////////////////////////////
/// Get the state of a key
////////////////////////////////////////////////////////////
bool sfInput::IsKeyDown(sfKey::Code Key) const
{
    return myKeys[Key];
}


////////////////////////////////////////////////////////////
/// Get the state of a mouse button
////////////////////////////////////////////////////////////
bool sfInput::IsMouseButtonDown(sfMouse::Button Button) const
{
    switch (Button)
    {
        case sfMouse::Left :   return myMouseButtons[0];
        case sfMouse::Right :  return myMouseButtons[1];
        case sfMouse::Middle : return myMouseButtons[2];
    }

    return false;
}


////////////////////////////////////////////////////////////
/// Get the state of a joystick button
////////////////////////////////////////////////////////////
bool sfInput::IsJoystickButtonDown(unsigned int JoyId, unsigned int Button) const
{
    if ((JoyId < 2) && (Button < 8))
        return myJoystickButtons[JoyId][Button];
    else
        return false;
}


////////////////////////////////////////////////////////////
/// Get the mouse left position
////////////////////////////////////////////////////////////
unsigned int sfInput::GetMouseX() const
{
    return myMouseX;
}


////////////////////////////////////////////////////////////
/// Get the mouse top position
////////////////////////////////////////////////////////////
unsigned int sfInput::GetMouseY() const
{
    return myMouseY;
}


////////////////////////////////////////////////////////////
/// Get the joystick X position
////////////////////////////////////////////////////////////
int sfInput::GetJoystickX(unsigned int JoyId) const
{
    return myJoystickX[JoyId];
}


////////////////////////////////////////////////////////////
/// Get the joystick Y position
////////////////////////////////////////////////////////////
int sfInput::GetJoystickY(unsigned int JoyId) const
{
    return myJoystickY[JoyId];
}


////////////////////////////////////////////////////////////
/// Get the joystick Z position
////////////////////////////////////////////////////////////
int sfInput::GetJoystickZ(unsigned int JoyId) const
{
    return myJoystickZ[JoyId];
}


////////////////////////////////////////////////////////////
/// /see sfWindowListener::OnEvent
////////////////////////////////////////////////////////////
void sfInput::OnEvent(const sfEvent& Event)
{
    switch (Event.Type)
    {
        // Key events
        case sfEvent::KeyPressed :  myKeys[Event.Key.Code] = true;  break;
        case sfEvent::KeyReleased : myKeys[Event.Key.Code] = false; break;

        // Mouse pressed event
        case sfEvent::MouseButtonPressed :
            switch (Event.Mouse.Buttons)
            {
                case sfMouse::Left :   myMouseButtons[0] = true; break;
                case sfMouse::Right :  myMouseButtons[1] = true; break;
                case sfMouse::Middle : myMouseButtons[2] = true; break;
            }
            break;

        // Mouse released event
        case sfEvent::MouseButtonReleased :
            switch (Event.Mouse.Buttons)
            {
                case sfMouse::Left :   myMouseButtons[0] = false; break;
                case sfMouse::Right :  myMouseButtons[1] = false; break;
                case sfMouse::Middle : myMouseButtons[2] = false; break;
            }
            break;

        // Mouse move event
        case sfEvent::MouseMove :
            myMouseX = Event.Mouse.X;
            myMouseY = Event.Mouse.Y;
            break;

        // Joystick button events
        case sfEvent::JoystickButtonPressed :  myJoystickButtons[Event.Joystick.JoystickId][Event.Joystick.Button] = true;  break;
        case sfEvent::JoystickButtonReleased : myJoystickButtons[Event.Joystick.JoystickId][Event.Joystick.Button] = false; break;

        // Joystick move event
        case sfEvent::JoystickMove :
            myJoystickX[Event.Joystick.JoystickId] = Event.Joystick.X;
            myJoystickY[Event.Joystick.JoystickId] = Event.Joystick.Y;
            myJoystickZ[Event.Joystick.JoystickId] = Event.Joystick.Z;
            break;

        default :
            break;
    }
}
