////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Window/Unix/WindowImplUnix.hpp>
#include <SFML/Window/Unix/glxext.h>
#include <X11/keysym.h>
#include <X11/extensions/Xrandr.h>
#include <linux/joystick.h>
#include <fcntl.h>
#include <iostream>
#include <sstream>


namespace sf_private
{
////////////////////////////////////////////////////////////
// Static member data
////////////////////////////////////////////////////////////
Display*          sfWindowImplUnix::ourDisplay          = NULL;
int               sfWindowImplUnix::ourScreen           = 0;
sfWindowImplUnix* sfWindowImplUnix::ourDummyWindow      = NULL;
sfWindowImplUnix* sfWindowImplUnix::ourFullscreenWindow = NULL;
unsigned int      sfWindowImplUnix::ourWindowsCount     = 0;
unsigned long     sfWindowImplUnix::ourEventMask        = FocusChangeMask | ButtonPressMask | ButtonReleaseMask | ButtonMotionMask |
                                                          PointerMotionMask | KeyPressMask | KeyReleaseMask | StructureNotifyMask;


////////////////////////////////////////////////////////////
/// Default constructor
/// (creates a dummy window to provide a valid OpenGL context)
////////////////////////////////////////////////////////////
sfWindowImplUnix::sfWindowImplUnix() :
myWindow      (0),
myIsExternal  (false),
myGLContext   (NULL),
myAtomClose   (0),
myOldVideoMode(-1),
myHiddenCursor(0)
{
    // Open the display at first call
    if (!OpenDisplay())
        return;

    // Use small dimensions
    myWidth  = 1;
    myHeight = 1;

    // Create the rendering context
    XVisualInfo Visual;
    if (!CreateContext(sfVideoMode(myWidth, myHeight, 32), false, Visual))
        return;

    // Create a new color map with the chosen visual
    Colormap ColMap = XCreateColormap(ourDisplay, RootWindow(ourDisplay, ourScreen), Visual.visual, AllocNone);

    // Define the window attributes
    XSetWindowAttributes Attributes;
    Attributes.colormap = ColMap;

    // Create a dummy window (disabled and hidden)
    myWindow = XCreateWindow(ourDisplay,
                             RootWindow(ourDisplay, ourScreen),
                             0, 0,
                             myWidth, myHeight,
                             0,
                             Visual.depth,
                             InputOutput,
                             Visual.visual,
                             CWColormap, &Attributes);

    // Save as the dummy window
    if (!ourDummyWindow)
        ourDummyWindow = this;

    // No joysticks
    myJoysticks[0] = myJoysticks[1] = -1;

    // Set our context as the current OpenGL context for rendering
    SetCurrent();
}


////////////////////////////////////////////////////////////
/// Create the window implementation from an existing control
////////////////////////////////////////////////////////////
sfWindowImplUnix::sfWindowImplUnix(sfWindowHandle Handle) :
myWindow      (0),
myIsExternal  (true),
myGLContext   (NULL),
myAtomClose   (0),
myOldVideoMode(-1),
myHiddenCursor(0)
{
    // Open the display at first call
    if (!OpenDisplay())
        return;

    // Save the window handle
    myWindow = Handle;

    if (myWindow)
    {
        // Get the window size
        XWindowAttributes WindowAttributes;
        if (XGetWindowAttributes(ourDisplay, myWindow, &WindowAttributes) == 0)
        {
            std::cerr << "Failed to get the window attributes" << std::endl;
            return;
        }
        myWidth  = WindowAttributes.width;
        myHeight = WindowAttributes.height;

        // Setup the visual infos to match
        XVisualInfo Template;
        Template.depth     = WindowAttributes.depth;
        Template.visualid  = XVisualIDFromVisual(WindowAttributes.visual);
        unsigned long Mask = VisualDepthMask | VisualIDMask;

        // Create the rendering context
        sfVideoMode Mode = sfVideoMode::GetDesktopMode();
        Mode.Width  = myWidth;
        Mode.Height = myHeight;
        XVisualInfo Visual;
        if (!CreateContext(Mode, false, Visual, Template, Mask))
            return;

        // Create a new color map with the chosen visual
        Colormap ColMap = XCreateColormap(ourDisplay, RootWindow(ourDisplay, ourScreen), Visual.visual, AllocNone);
        XSetWindowColormap(ourDisplay, myWindow, ColMap);

        // Make sure the window is listening to all the requiered events
        XSelectInput(ourDisplay, myWindow, ourEventMask & ~ButtonPressMask);

        // Do some common initializations
        Initialize();
    }
}


////////////////////////////////////////////////////////////
/// Create the window implementation
////////////////////////////////////////////////////////////
sfWindowImplUnix::sfWindowImplUnix(sfVideoMode Mode, const std::string& Title, bool Fullscreen) :
myWindow      (0),
myIsExternal  (false),
myGLContext   (NULL),
myAtomClose   (0),
myOldVideoMode(-1),
myHiddenCursor(0)
{
    // Open the display at first call
    if (!OpenDisplay())
        return;

    // Compute position and size
    int Left, Top;
    if (!Fullscreen)
    {
        Left = (DisplayWidth(ourDisplay, ourScreen)  - Mode.Width)  / 2;
        Top  = (DisplayHeight(ourDisplay, ourScreen) - Mode.Height) / 2;
    }
    else
    {
        Left = 0;
        Top  = 0;
    }
    int Width  = myWidth  = Mode.Width;
    int Height = myHeight = Mode.Height;

    // Create the rendering context
    XVisualInfo Visual;
    if (!CreateContext(Mode, Fullscreen, Visual))
        return;

    // Create a new color map with the chosen visual
    Colormap ColMap = XCreateColormap(ourDisplay, RootWindow(ourDisplay, ourScreen), Visual.visual, AllocNone);

    // Define the window attributes
    XSetWindowAttributes Attributes;
    Attributes.event_mask        = ourEventMask;
    Attributes.colormap          = ColMap;
    Attributes.override_redirect = Fullscreen;

    // Create the window
    myWindow = XCreateWindow(ourDisplay,
                             RootWindow(ourDisplay, ourScreen),
                             Left, Top,
                             Width, Height,
                             0,
                             Visual.depth,
                             InputOutput,
                             Visual.visual,
                             CWEventMask | CWColormap | CWOverrideRedirect, &Attributes);
    if (!myWindow)
    {
        std::cerr << "Failed to create window" << std::endl;
        return;
    }

    // Set the window's name
    XStoreName(ourDisplay, myWindow, Title.c_str());

    // Do some common initializations
    Initialize();

    // In fullscreen mode, we must grab keyboard and mouse inputs
    if (Fullscreen)
    {
        XGrabPointer(ourDisplay, myWindow, true, 0, GrabModeAsync, GrabModeAsync, myWindow, None, CurrentTime);
        XGrabKeyboard(ourDisplay, myWindow, true, GrabModeAsync, GrabModeAsync, CurrentTime);
    }
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
sfWindowImplUnix::~sfWindowImplUnix()
{
    // Cleanup graphical resources
    CleanUp();

    // Destroy the window
    if (myWindow && !myIsExternal)
        XDestroyWindow(ourDisplay, myWindow);

    // Close the display
    CloseDisplay();
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::Display
////////////////////////////////////////////////////////////
void sfWindowImplUnix::Display()
{
    if (myWindow && myGLContext)
        glXSwapBuffers(ourDisplay, myWindow);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::ProcessEvents
////////////////////////////////////////////////////////////
void sfWindowImplUnix::ProcessEvents()
{
    // Get the number of events in the queue
    int NbEvents = XPending(ourDisplay);

    // Process the events
    XEvent Event;
    for (int i = 0; i < NbEvents; ++i)
    {
        XNextEvent(ourDisplay, &Event);
        if (Event.xany.window == myWindow)
            ProcessEvent(Event);
        else
            XPutBackEvent(ourDisplay, &Event);
    }

    // Read the joysticks state
    if (myJoysticks[0] > 0) ReadJoyState(0);
    if (myJoysticks[1] > 0) ReadJoyState(1);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::MakeCurrent
////////////////////////////////////////////////////////////
void sfWindowImplUnix::MakeCurrent() const
{
    if (myWindow && myGLContext)
        glXMakeCurrent(ourDisplay, myWindow, myGLContext);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::UseVerticalSync
////////////////////////////////////////////////////////////
void sfWindowImplUnix::UseVerticalSync(bool Enabled)
{
    const GLubyte* ProcAddress = reinterpret_cast<const GLubyte*>("glXSwapIntervalSGI");
    PFNGLXSWAPINTERVALSGIPROC glXSwapIntervalSGI = reinterpret_cast<PFNGLXSWAPINTERVALSGIPROC>(glXGetProcAddress(ProcAddress));
    if (glXSwapIntervalSGI)
        glXSwapIntervalSGI(Enabled ? 1 : 0);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::ShowMouseCursor
////////////////////////////////////////////////////////////
void sfWindowImplUnix::ShowMouseCursor(bool Show)
{
    XDefineCursor(ourDisplay, myWindow, Show ? None : myHiddenCursor);
}


////////////////////////////////////////////////////////////
/// Create the OpenGL rendering context
////////////////////////////////////////////////////////////
bool sfWindowImplUnix::CreateContext(sfVideoMode Mode, bool Fullscreen, XVisualInfo& ChosenVisual, XVisualInfo Template, unsigned long Mask)
{
    // Fullscreen mode
    if (Fullscreen)
    {
        // Check if the XRandR extension is present
        int Version;
        if (XQueryExtension(ourDisplay, "RANDR", &Version, &Version, &Version))
        {
            // Get the current configuration
            XRRScreenConfiguration* Config = XRRGetScreenInfo(ourDisplay, RootWindow(ourDisplay, ourScreen));
            if (Config)
            {
                // Get the current rotation
                Rotation CurrentRotation;
                myOldVideoMode = XRRConfigCurrentConfiguration(Config, &CurrentRotation);

                // Get the available screen sizes
                int NbSizes;
                XRRScreenSize* Sizes = XRRConfigSizes(Config, &NbSizes);
                if (Sizes && (NbSizes > 0))
                {
                    // Search a matching size
                    for (int i = 0; i < NbSizes; ++i)
                    {
                        if ((Sizes[i].width == static_cast<int>(Mode.Width)) && (Sizes[i].height == static_cast<int>(Mode.Height)))
                        {
                            // Switch to fullscreen mode
                            XRRSetScreenConfig(ourDisplay, Config, RootWindow(ourDisplay, ourScreen), i, CurrentRotation, CurrentTime);

                            // Set "this" as the current fullscreen window
                            ourFullscreenWindow = this;
                            break;
                        }
                    }
                }

                // Free the configuration instance
                XRRFreeScreenConfigInfo(Config);
            }
            else
            {
                // Failed to get the screen configuration
                std::cerr << "Failed to get the current screen configuration for fullscreen mode, switching to windiw mode" << std::endl;
            }
        }
        else
        {
            // XRandr extension is not supported : we cannot use fullscreen mode
            std::cerr << "Fullscreen is not supported, switching to window mode" << std::endl;
        }
    }

    // Get all the visuals matching the template
    Template.screen = ourScreen;
    int NbVisuals = 0;
    XVisualInfo* Visuals = XGetVisualInfo(ourDisplay, Mask | VisualScreenMask, &Template, &NbVisuals);
    if (!Visuals || (NbVisuals == 0))
    {
        if (Visuals)
            XFree(Visuals);
        std::cerr << "There is no valid visual for the selected screen" << std::endl;
        return false;
    }

    // A helper structure for storing visual attributes
    struct VisualAttributes
    {
        int RGBA, DoubleBuffer, Red, Green, Blue, Alpha, Depth, Stencil;
    };

    // Find the best visual
    VisualAttributes BestAttribs = {0, 0, 0, 0, 0, 0, 0, 0};
    XVisualInfo* Visual = &Visuals[0];
    for (int i = 0; i < NbVisuals; ++i)
    {
        // Get the visual attributes
        VisualAttributes Attribs = {0, 0, 0, 0, 0, 0, 0, 0};
        glXGetConfig(ourDisplay, &Visuals[i], GLX_RGBA,         &Attribs.RGBA);
        glXGetConfig(ourDisplay, &Visuals[i], GLX_DOUBLEBUFFER, &Attribs.DoubleBuffer); 
        glXGetConfig(ourDisplay, &Visuals[i], GLX_RED_SIZE,     &Attribs.Red);
        glXGetConfig(ourDisplay, &Visuals[i], GLX_GREEN_SIZE,   &Attribs.Green); 
        glXGetConfig(ourDisplay, &Visuals[i], GLX_BLUE_SIZE,    &Attribs.Blue); 
        glXGetConfig(ourDisplay, &Visuals[i], GLX_ALPHA_SIZE,   &Attribs.Alpha); 
        glXGetConfig(ourDisplay, &Visuals[i], GLX_DEPTH_SIZE,   &Attribs.Depth);        
        glXGetConfig(ourDisplay, &Visuals[i], GLX_STENCIL_SIZE, &Attribs.Stencil);

        // Compare with the current best
        if ((BestAttribs.RGBA    < Attribs.RGBA)    || (BestAttribs.DoubleBuffer < Attribs.DoubleBuffer) ||
            (BestAttribs.Red     < Attribs.Red)     || (BestAttribs.Green        < Attribs.Green)        ||
            (BestAttribs.Blue    < Attribs.Blue)    || (BestAttribs.Alpha        < Attribs.Alpha)        ||
            (BestAttribs.Stencil < Attribs.Stencil) || (BestAttribs.Depth        < Attribs.Depth))
        {
            BestAttribs = Attribs;
            Visual = &Visuals[i];
        }
    }

    // Create the OpenGL context
    myGLContext = glXCreateContext(ourDisplay, Visual, glXGetCurrentContext(), true);
    if (myGLContext == NULL)
    {
        std::cerr << "Failed to create an OpenGL context for this window" << std::endl;
        return false;
    }

    // Extract the depth and stencil bits from the chosen format
    int Depth, Stencil;
    glXGetConfig(ourDisplay, Visual, GLX_DEPTH_SIZE,   &Depth);
    glXGetConfig(ourDisplay, Visual, GLX_STENCIL_SIZE, &Stencil);
    myDepthBits   = static_cast<unsigned int>(Depth);
    myStencilBits = static_cast<unsigned int>(Stencil);

    // Assign the chosen visual, and free the temporary visuals array
    ChosenVisual = *Visual;
    XFree(Visuals);

    return true;
}


////////////////////////////////////////////////////////////
/// Do some common initializations after the window has been created
////////////////////////////////////////////////////////////
void sfWindowImplUnix::Initialize()
{
    // Get the atom defining the close event
    myAtomClose = XInternAtom(ourDisplay, "WM_DELETE_WINDOW", false);
    XSetWMProtocols(ourDisplay, myWindow, &myAtomClose, 1);

    // Show the window
    XMapWindow(ourDisplay, myWindow);
    XFlush(ourDisplay);

    // Setup the joysticks
    SetupJoystick(0);
    SetupJoystick(1);

    // Create the hiden cursor
    CreateHiddenCursor();

    // Set our context as the current OpenGL context for rendering
    SetCurrent();

    // Flush the commands queue
    XFlush(ourDisplay);
}


////////////////////////////////////////////////////////////
/// Setup a joystick
////////////////////////////////////////////////////////////
void sfWindowImplUnix::SetupJoystick(unsigned int Id)
{
    // Open the joystick handle
    std::ostringstream oss;
    oss << "/dev/input/js" << Id;
    myJoysticks[Id] = open(oss.str().c_str(), O_RDONLY);
    if (myJoysticks[Id] > 0)
    {
        // Use non-blocking mode
        fcntl(myJoysticks[Id], F_SETFL, O_NONBLOCK);
    }
}


////////////////////////////////////////////////////////////
/// Read the state of a given joystick, and triggers the
/// appropriate events
////////////////////////////////////////////////////////////
void sfWindowImplUnix::ReadJoyState(unsigned int Id)
{
    js_event JoyState;
    while (read(myJoysticks[Id], &JoyState, sizeof(JoyState)) > 0)
    {
        switch (JoyState.type & ~JS_EVENT_INIT)
        {
            // An axis has been moved
            case JS_EVENT_AXIS :
            {
                sfEvent Event;
                Event.Type                = sfEvent::JoystickMove;
                Event.Joystick.JoystickId = Id;
                if      (JoyState.number == 0) Event.Joystick.X = JoyState.value * 100 / 32767;
                else if (JoyState.number == 1) Event.Joystick.Y = JoyState.value * 100 / 32767;
                else if (JoyState.number == 2) Event.Joystick.Z = JoyState.value * 100 / 32767;

                SendEvent(Event);
                break;
            }

            // A button has been pressed
            case JS_EVENT_BUTTON :
            {
                sfEvent Event;
                Event.Type                = JoyState.value ? sfEvent::JoystickButtonPressed : sfEvent::JoystickButtonReleased;
                Event.Joystick.JoystickId = Id;
                Event.Joystick.Button     = JoyState.number;

                SendEvent(Event);
                break;
            }
        }
    }
}


////////////////////////////////////////////////////////////
/// Create a transparent mouse cursor
////////////////////////////////////////////////////////////
void sfWindowImplUnix::CreateHiddenCursor()
{
    // Create the cursor's pixmap (1x1 pixels)
	Pixmap CursorPixmap = XCreatePixmap(ourDisplay, myWindow, 1, 1, 1);
    GC GraphicsContext = XCreateGC(ourDisplay, CursorPixmap, 0, NULL);
    XDrawPoint(ourDisplay, CursorPixmap, GraphicsContext, 0, 0);
    XFreeGC(ourDisplay, GraphicsContext);

    // Create the cursor, using the pixmap as both the shape and the mask of the cursor
	XColor Color;
	Color.flags = DoRed | DoGreen | DoBlue;
	Color.red = Color.blue = Color.green = 0;
    myHiddenCursor = XCreatePixmapCursor(ourDisplay, CursorPixmap, CursorPixmap, &Color, &Color, 0, 0);

    // We don't need the pixmap any longer, free it
	XFreePixmap(ourDisplay, CursorPixmap);
}


////////////////////////////////////////////////////////////
/// Cleanup graphical resources attached to the window
////////////////////////////////////////////////////////////
void sfWindowImplUnix::CleanUp()
{
    // Restore previous video mode (in case we were running in fullscreen)
    if (ourFullscreenWindow == this)
    {
        // Get current screen info
        XRRScreenConfiguration* Config = XRRGetScreenInfo(ourDisplay, RootWindow(ourDisplay, ourScreen));
        if (Config) 
        {
            // Get the current rotation
            Rotation CurrentRotation;
            XRRConfigCurrentConfiguration(Config, &CurrentRotation);

            // Reset the video mode
            XRRSetScreenConfig(ourDisplay, Config, RootWindow(ourDisplay, ourScreen), myOldVideoMode, CurrentRotation, CurrentTime);

            // Free the configuration instance
            XRRFreeScreenConfigInfo(Config);
        } 

        // Reset the fullscreen window
        ourFullscreenWindow = NULL;
    }

    // Unhide the mouse cursor (in case it was hidden)
    ShowMouseCursor(true);

    // Close the joysticks handles
    if (myJoysticks[0] > 0) close(myJoysticks[0]);
    if (myJoysticks[1] > 0) close(myJoysticks[1]);

    // Destroy the OpenGL context
    if (myGLContext)
    {
        // If this is not the dummy window, we must set it as the valid rendering context to avoid a crash with next OpenGL command
        if (this != ourDummyWindow)
        {
            if (ourDummyWindow)
                ourDummyWindow->SetCurrent();
        }
        else
        {
            ourDummyWindow = NULL;
            glXMakeCurrent(ourDisplay, 0, NULL);
        }

        // Now that our context is no more the current one, we can destroy it
        glXDestroyContext(ourDisplay, myGLContext);
        myGLContext = NULL;
    }
}


////////////////////////////////////////////////////////////
/// Process an incoming event from the window
////////////////////////////////////////////////////////////
void sfWindowImplUnix::ProcessEvent(XEvent WinEvent)
{
    switch (WinEvent.type)
    {
        // Destroy event
        case DestroyNotify :
        {
            // The window is about to be destroyed : we must cleanup resources
            CleanUp();
            break;
        }

        // Gain focus event
        case FocusIn :
        {
            sfEvent Event;
            Event.Type = sfEvent::GainedFocus;
            SendEvent(Event);
            break;
        }

        // Lost focus event
        case FocusOut :
        {
            sfEvent Event;
            Event.Type = sfEvent::LostFocus;
            SendEvent(Event);
            break;
        }

        // Resize event
        case ConfigureNotify :
        {
            if ((WinEvent.xconfigure.width != static_cast<int>(myWidth)) || (WinEvent.xconfigure.height != static_cast<int>(myHeight)))
            {
                sfEvent Event;
                Event.Type        = sfEvent::Resize;
                Event.Size.Width  = WinEvent.xconfigure.width;
                Event.Size.Height = WinEvent.xconfigure.height;
                SendEvent(Event);
            }
            break;
        }

        // Close event
        case ClientMessage :
        {
            if ((WinEvent.xclient.format == 32) && (WinEvent.xclient.data.l[0]) == static_cast<long>(myAtomClose))  
            {
                sfEvent Event;
                Event.Type = sfEvent::Close;
                SendEvent(Event);
                CleanUp();
            }
            break;
        }

        // Key down event
        case KeyPress :
        {
            // Get the keysym of the key that has been pressed
            char Buffer[32];
            KeySym Sym;
            XLookupString(&WinEvent.xkey, Buffer, 32, &Sym, NULL);

            // Fill the event parameters
            sfEvent Event;
            Event.Type        = sfEvent::KeyPressed;
            Event.Key.Code    = KeysymToSF(Sym);
            Event.Key.Alt     = WinEvent.xkey.state & Mod1Mask;
            Event.Key.Control = WinEvent.xkey.state & ControlMask;
            Event.Key.Shift   = WinEvent.xkey.state & ShiftMask;

            // TODO : Generate a TextEntered event

            SendEvent(Event);
            break;
        }

        // Key up event
        case KeyRelease :
        {
            // Get the keysym of the key that has been pressed
            char Buffer[32];
            KeySym Sym;
            XLookupString(&WinEvent.xkey, Buffer, 32, &Sym, NULL);

            // Fill the event parameters
            sfEvent Event;
            Event.Type        = sfEvent::KeyReleased;
            Event.Key.Code    = KeysymToSF(Sym);
            Event.Key.Alt     = WinEvent.xkey.state & Mod1Mask;
            Event.Key.Control = WinEvent.xkey.state & ControlMask;
            Event.Key.Shift   = WinEvent.xkey.state & ShiftMask;

            // TODO : Generate a TextEntered event

            SendEvent(Event);
            break;
        }

        // Mouse button pressed
        case ButtonPress :
        {
            sfEvent Event;
            Event.Type    = sfEvent::MouseButtonPressed;
            Event.Mouse.X = WinEvent.xbutton.x;
            Event.Mouse.Y = WinEvent.xbutton.y;
            switch (WinEvent.xbutton.button)
            {
                case Button1 : Event.Mouse.Buttons = sfMouse::Left;   break;
                case Button2 : Event.Mouse.Buttons = sfMouse::Middle; break;
                case Button3 : Event.Mouse.Buttons = sfMouse::Right;  break;
            }
            SendEvent(Event);
            break;
        }

        // Mouse button released
        case ButtonRelease :
        {
            sfEvent Event;
            Event.Type    = sfEvent::MouseButtonReleased;
            Event.Mouse.X = WinEvent.xbutton.x;
            Event.Mouse.Y = WinEvent.xbutton.y;
            switch (WinEvent.xbutton.button)
            {
                case Button1 : Event.Mouse.Buttons = sfMouse::Left;   break;
                case Button2 : Event.Mouse.Buttons = sfMouse::Middle; break;
                case Button3 : Event.Mouse.Buttons = sfMouse::Right;  break;
            }
            SendEvent(Event);
            break;
        }

        // Mouse moved
        case MotionNotify :
        {
            sfEvent Event;
            Event.Type    = sfEvent::MouseMove;
            Event.Mouse.X = WinEvent.xmotion.x;
            Event.Mouse.Y = WinEvent.xmotion.y;
            Event.Mouse.Buttons = 0;
            if (WinEvent.xmotion.state & Button1Mask) Event.Mouse.Buttons |= sfMouse::Left;
            if (WinEvent.xmotion.state & Button2Mask) Event.Mouse.Buttons |= sfMouse::Middle;
            if (WinEvent.xmotion.state & Button3Mask) Event.Mouse.Buttons |= sfMouse::Right;
            SendEvent(Event);
            break;
        }
    }
}


////////////////////////////////////////////////////////////
/// Convert a X11 keysym to SFML key code
////////////////////////////////////////////////////////////
sfKey::Code sfWindowImplUnix::KeysymToSF(KeySym Sym)
{
    // First convert to uppercase (to avoid dealing with two different keysyms for the same key)
    KeySym Lower, Key;
    XConvertCase(Sym, &Lower, &Key);

    switch (Key)
    {
        case XK_Escape :      return sfKey::Escape;
        case XK_space :       return sfKey::Space;
        case XK_Return :      return sfKey::Return;
        case XK_BackSpace :   return sfKey::Back;
        case XK_Tab :         return sfKey::Tab;
        case XK_Prior :       return sfKey::PageUp;
        case XK_Next :        return sfKey::PageDown;
        case XK_End :         return sfKey::End;
        case XK_Home :        return sfKey::Home;
        case XK_Insert :      return sfKey::Insert;
        case XK_Delete :      return sfKey::Delete;
        case XK_KP_Add :      return sfKey::Add;
        case XK_KP_Subtract : return sfKey::Subtract;
        case XK_KP_Multiply : return sfKey::Multiply;
        case XK_KP_Divide :   return sfKey::Divide;
        case XK_F1 :          return sfKey::F1;
        case XK_F2 :          return sfKey::F2;
        case XK_F3 :          return sfKey::F3;
        case XK_F4 :          return sfKey::F4;
        case XK_F5 :          return sfKey::F5;
        case XK_F6 :          return sfKey::F6;
        case XK_F7 :          return sfKey::F7;
        case XK_F8 :          return sfKey::F8;
        case XK_F9 :          return sfKey::F9;
        case XK_F10 :         return sfKey::F10;
        case XK_F11 :         return sfKey::F11;
        case XK_F12 :         return sfKey::F12;
        case XK_F13 :         return sfKey::F13;
        case XK_F14 :         return sfKey::F14;
        case XK_F15 :         return sfKey::F15;
        case XK_Left :        return sfKey::Left;
        case XK_Right :       return sfKey::Right;
        case XK_Up :          return sfKey::Up;
        case XK_Down :        return sfKey::Down;
        case XK_KP_0 :        return sfKey::Numpad0;
        case XK_KP_1 :        return sfKey::Numpad1;
        case XK_KP_2 :        return sfKey::Numpad2;
        case XK_KP_3 :        return sfKey::Numpad3;
        case XK_KP_4 :        return sfKey::Numpad4;
        case XK_KP_5 :        return sfKey::Numpad5;
        case XK_KP_6 :        return sfKey::Numpad6;
        case XK_KP_7 :        return sfKey::Numpad7;
        case XK_KP_8 :        return sfKey::Numpad8;
        case XK_KP_9 :        return sfKey::Numpad9;
        case XK_A :           return sfKey::A;
        case XK_Z :           return sfKey::Z;
        case XK_E :           return sfKey::E;
        case XK_R :           return sfKey::R;
        case XK_T :           return sfKey::T;
        case XK_Y :           return sfKey::Y;
        case XK_U :           return sfKey::U;
        case XK_I :           return sfKey::I;
        case XK_O :           return sfKey::O;
        case XK_P :           return sfKey::P;
        case XK_Q :           return sfKey::Q;
        case XK_S :           return sfKey::S;
        case XK_D :           return sfKey::D;
        case XK_F :           return sfKey::F;
        case XK_G :           return sfKey::G;
        case XK_H :           return sfKey::H;
        case XK_J :           return sfKey::J;
        case XK_K :           return sfKey::K;
        case XK_L :           return sfKey::L;
        case XK_M :           return sfKey::M;
        case XK_W :           return sfKey::W;
        case XK_X :           return sfKey::X;
        case XK_C :           return sfKey::C;
        case XK_V :           return sfKey::V;
        case XK_B :           return sfKey::B;
        case XK_N :           return sfKey::N;
        case XK_0 :           return sfKey::Num0;
        case XK_1 :           return sfKey::Num1;
        case XK_2 :           return sfKey::Num2;
        case XK_3 :           return sfKey::Num3;
        case XK_4 :           return sfKey::Num4;
        case XK_5 :           return sfKey::Num5;
        case XK_6 :           return sfKey::Num6;
        case XK_7 :           return sfKey::Num7;
        case XK_8 :           return sfKey::Num8;
        case XK_9 :           return sfKey::Num9;
    }

    return sfKey::Code(0);
}


////////////////////////////////////////////////////////////
/// Open the display (if not already done)
////////////////////////////////////////////////////////////
bool sfWindowImplUnix::OpenDisplay(bool AddWindow)
{
    // If no display has been opened yet, open it
    if (ourDisplay == NULL)
    {
        ourDisplay = XOpenDisplay(NULL);
        if (ourDisplay)
            ourScreen = DefaultScreen(ourDisplay);
        else
            std::cerr << "Failed to open a connection with the X server" << std::endl;
    }

    // Increase the number of windows
    if (AddWindow)
        ourWindowsCount++;

    return ourDisplay != NULL;
}


////////////////////////////////////////////////////////////
/// Close the display
////////////////////////////////////////////////////////////
void sfWindowImplUnix::CloseDisplay()
{
    // Decrease the number of windows
    ourWindowsCount--;

    // If all windows have been destroyed, then we can close the display
    if (ourWindowsCount == 0)
    {
        XCloseDisplay(ourDisplay);
        ourDisplay = NULL;
    }
}

} // namespace sf_private
