////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

#ifndef SFML_WINDOWIMPLUNIX_HPP
#define SFML_WINDOWIMPLUNIX_HPP

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Window/Event.hpp>
#include <SFML/Window/WindowImpl.hpp>
#include <X11/Xlib.h>
#include <GL/glx.h>
#include <set>
#include <string>


namespace sf_private
{
class sfVideoModeSupport;

////////////////////////////////////////////////////////////
/// sfWindowImplUnix is the Unix (X11) implementation of sfWindowImpl
////////////////////////////////////////////////////////////
class sfWindowImplUnix : public sfWindowImpl
{
public :

    ////////////////////////////////////////////////////////////
    /// Default constructor
    /// (creates a dummy window to provide a valid OpenGL context)
    ///
    ////////////////////////////////////////////////////////////
    sfWindowImplUnix();

    ////////////////////////////////////////////////////////////
    /// Construct the window implementation from an existing control
    ///
    /// \param Handle : Platform-specific handle of the control
    ///
    ////////////////////////////////////////////////////////////
    sfWindowImplUnix(sfWindowHandle Handle);

    ////////////////////////////////////////////////////////////
    /// Create the window implementation
    ///
    /// \param Mode :       Video mode to use
    /// \param Title :      Title of the window
    /// \param Fullscreen : True to set fullscreen, false to stay in windowed mode
    ///
    ////////////////////////////////////////////////////////////
    sfWindowImplUnix(sfVideoMode Mode, const std::string& Title, bool Fullscreen);

    ////////////////////////////////////////////////////////////
    /// Destructor
    ///
    ////////////////////////////////////////////////////////////
    ~sfWindowImplUnix();

private :

    friend class sfVideoModeSupport;

    ////////////////////////////////////////////////////////////
    /// /see sfWindowImpl::Display
    ///
    ////////////////////////////////////////////////////////////
    virtual void Display();

    ////////////////////////////////////////////////////////////
    /// /see sfWindowImpl::ProcessEvents
    ///
    ////////////////////////////////////////////////////////////
    virtual void ProcessEvents();

    ////////////////////////////////////////////////////////////
    /// /see sfWindowImpl::MakeCurrent
    ///
    ////////////////////////////////////////////////////////////
    virtual void MakeCurrent() const;

    ////////////////////////////////////////////////////////////
    /// /see sfWindowImpl::UseVerticalSync
    ///
    ////////////////////////////////////////////////////////////
    virtual void UseVerticalSync(bool Enabled);

    ////////////////////////////////////////////////////////////
    /// /see sfWindowImpl::ShowMouseCursor
    ///
    ////////////////////////////////////////////////////////////
    virtual void ShowMouseCursor(bool Show);

    ////////////////////////////////////////////////////////////
    /// Create the OpenGL rendering context
    ///
    /// \param Mode :         Video mode to use
    /// \param Fullscreen :   True to set fullscreen, false to stay in windowed mode
    /// \param ChosenVisual : Visual that has been chosen for creating the contexte
    /// \param Template :     Visual infos to match
    /// \param Mask :         Visual attributes to check in Template
    ///
    /// \return True on success, false on error
    ///
    ////////////////////////////////////////////////////////////
    bool CreateContext(sfVideoMode Mode, bool Fullscreen, XVisualInfo& ChosenVisual, XVisualInfo Template = XVisualInfo(), unsigned long Mask = 0);

    ////////////////////////////////////////////////////////////
    /// Do some common initializations after the window has been created
    ///
    ////////////////////////////////////////////////////////////
    void Initialize();

    ////////////////////////////////////////////////////////////
    /// Setup a joystick
    ///
    /// \param Id : Identifier of the joystick to setup
    ///
    ////////////////////////////////////////////////////////////
    void SetupJoystick(unsigned int Id);

    ////////////////////////////////////////////////////////////
    /// Read the state of a given joystick, and triggers the
    /// appropriate events
    ///
    /// \param Id : Identifier of the joystick to read
    ///
    ///////////////////////////////////////////////////////////
    void ReadJoyState(unsigned int Id);

    ////////////////////////////////////////////////////////////
    /// Create a transparent mouse cursor
    ///
    ////////////////////////////////////////////////////////////
    void CreateHiddenCursor();

    ////////////////////////////////////////////////////////////
    /// Cleanup graphical resources attached to the window
    ///
    ////////////////////////////////////////////////////////////
    void CleanUp();

    ////////////////////////////////////////////////////////////
    /// Process an incoming event from the window
    ///
    /// \param WinEvent : Event which has been received
    ///
    ////////////////////////////////////////////////////////////
    void ProcessEvent(XEvent WinEvent);

    ////////////////////////////////////////////////////////////
    /// Convert a X11 keysym to SFML key code
    ///
    /// \param Sym : keysym to convert
    ///
    /// \return Corrsponding SFML key code
    ///
    ////////////////////////////////////////////////////////////
    static sfKey::Code KeysymToSF(KeySym Sym);

    ////////////////////////////////////////////////////////////
    /// Open the display (if not already done)
    ///
    /// \param AddWindow : Tell whether or not we must increase the windows count
    ///
    /// \return True if the display is properly opened
    ///
    ////////////////////////////////////////////////////////////
    static bool OpenDisplay(bool AddWindow = true);

    ////////////////////////////////////////////////////////////
    /// Close the display
    ///
    ////////////////////////////////////////////////////////////
    static void CloseDisplay();

    ////////////////////////////////////////////////////////////
    // Static member data
    ////////////////////////////////////////////////////////////
    static ::Display*        ourDisplay;          ///< Current opened display
    static int               ourScreen;           ///< Default screen on the opened display
    static sfWindowImplUnix* ourDummyWindow;      ///< Pointer to the dummy window
    static sfWindowImplUnix* ourFullscreenWindow; ///< Keep track of the active fullscreen window
    static unsigned int      ourWindowsCount;     ///< Number of windows created
    static unsigned long     ourEventMask;        ///< Mask defining the events that will be caught by our windows

    ////////////////////////////////////////////////////////////
    // Member data
    ////////////////////////////////////////////////////////////
    Window     myWindow;       ///< X11 structure defining our window
    bool       myIsExternal;   ///< Tell whether the window has been created externally or by SFML
    GLXContext myGLContext;    ///< OpenGL context attached to the window
    Atom       myAtomClose;    ///< Atom used to identify the close event
    int        myOldVideoMode; ///< Video mode in use before we switch to fullscreen
    Cursor     myHiddenCursor; ///< As X11 doesn't provide cursor hidding, we must create a transparent one
    int        myJoysticks[2]; ///< File handles connected to the joystick inputs
};

} // namespace sf_private

#endif // SFML_WINDOWIMPLX11_HPP
