////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Window/Win32/WindowImplWin32.hpp>
#include <GL/gl.h>
#include <SFML/Window/Win32/wglext.h>
#include <mmsystem.h>
#include <iostream>


namespace sf_private
{
////////////////////////////////////////////////////////////
// Static member data
////////////////////////////////////////////////////////////
unsigned int       sfWindowImplWin32::ourWindowCount      = 0;
const wchar_t*     sfWindowImplWin32::ourClassName        = L"SFML_Window";
sfWindowImplWin32* sfWindowImplWin32::ourDummyWindow      = NULL;
sfWindowImplWin32* sfWindowImplWin32::ourFullscreenWindow = NULL;


////////////////////////////////////////////////////////////
/// Default constructor
/// (creates a dummy window to provide a valid OpenGL context)
////////////////////////////////////////////////////////////
sfWindowImplWin32::sfWindowImplWin32() :
myHandle  (NULL),
myCallback(0)
{
    // Register the window class at first call
    if (ourWindowCount == 0)
        RegisterWindowClass();

    // Use small dimensions
    myWidth  = 1;
    myHeight = 1;

    // Create a dummy window (disabled and hidden)
    myHandle = CreateWindow(ourClassName, TEXT(""), WS_POPUP | WS_DISABLED, 0, 0, myWidth, myHeight, NULL, NULL, GetModuleHandle(NULL), NULL);
    ShowWindow(myHandle, SW_HIDE);

    // Create the rendering context
    if (myHandle)
        CreateContext(sfVideoMode(myWidth, myHeight, 32), false);

    // Save as the dummy window
    if (!ourDummyWindow)
        ourDummyWindow = this;
}


////////////////////////////////////////////////////////////
/// Create the window implementation from an existing control
////////////////////////////////////////////////////////////
sfWindowImplWin32::sfWindowImplWin32(sfWindowHandle Handle) :
myCallback(0)
{
    // Save window handle
    myHandle = static_cast<HWND>(Handle);

    if (myHandle)
    {
        // Get window client size
        RECT Rect;
        GetClientRect(myHandle, &Rect);
        myWidth  = Rect.right - Rect.left;
        myHeight = Rect.bottom - Rect.top;

        // Create the rendering context
        sfVideoMode Mode = sfVideoMode::GetDesktopMode();
        Mode.Width  = myWidth;
        Mode.Height = myHeight;
        CreateContext(Mode, false);

        // We change the event procedure of the control (it is important to save the old one)
        SetWindowLongPtr(myHandle, GWLP_USERDATA, reinterpret_cast<long>(this));
        myCallback = SetWindowLongPtr(myHandle, GWLP_WNDPROC, reinterpret_cast<long>(&sfWindowImplWin32::GlobalOnEvent));

        // Setup joysticks
        SetupJoysticks();
    }
}


////////////////////////////////////////////////////////////
/// Create the window implementation
////////////////////////////////////////////////////////////
sfWindowImplWin32::sfWindowImplWin32(sfVideoMode Mode, const std::string& Title, bool Fullscreen) :
myHandle  (NULL),
myCallback(0)
{
    // Register the window class at first call
    if (ourWindowCount == 0)
        RegisterWindowClass();

    // Compute position and size
    int Left   = (GetDeviceCaps(GetDC(NULL), HORZRES) - Mode.Width)  / 2;
    int Top    = (GetDeviceCaps(GetDC(NULL), VERTRES) - Mode.Height) / 2;
    int Width  = myWidth  = Mode.Width;
    int Height = myHeight = Mode.Height;

    // In windowed mode, adjust width and height so that window will have the requested client area
    if (!Fullscreen)
    {
        RECT Rect = {0, 0, Width, Height};
        AdjustWindowRect(&Rect, WS_OVERLAPPEDWINDOW | WS_VISIBLE, false);
        Width  = Rect.right - Rect.left;
        Height = Rect.bottom - Rect.top;
    }

    // Create the window
    wchar_t WTitle[256];
    int NbChars = MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, Title.c_str(), static_cast<int>(Title.size()), WTitle, sizeof(WTitle) / sizeof(*WTitle));
    WTitle[NbChars] = L'\0';
    myHandle = CreateWindow(ourClassName, WTitle, WS_OVERLAPPEDWINDOW | WS_VISIBLE, Left, Top, Width, Height, NULL, NULL, GetModuleHandle(NULL), this);

    // Create the rendering context
    if (myHandle)
        CreateContext(Mode, Fullscreen);

    // Increment window count
    ourWindowCount++;

    // Setup joysticks
    SetupJoysticks();
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
sfWindowImplWin32::~sfWindowImplWin32()
{
    if (!myCallback)
    {
        // Destroy the window
        if (myHandle)
            DestroyWindow(myHandle);

        // Decrement the window count
        ourWindowCount--;

        // Unregister window class if we were the last window
        if (ourWindowCount == 0)
            UnregisterClass(ourClassName, GetModuleHandle(NULL));
    }
    else
    {
        // The window is external : remove the hook on its message callback
        SetWindowLongPtr(myHandle, GWLP_WNDPROC, myCallback);
    }
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::Display
////////////////////////////////////////////////////////////
void sfWindowImplWin32::Display()
{
    if (myDeviceContext && myGLContext)
        SwapBuffers(myDeviceContext);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::ProcessEvents
////////////////////////////////////////////////////////////
void sfWindowImplWin32::ProcessEvents()
{
    // We update the window only if we own it
    if (!myCallback)
    {
        MSG Message;
        while (PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
        {
            TranslateMessage(&Message);
            DispatchMessage(&Message);
        }
    }
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::MakeCurrent
////////////////////////////////////////////////////////////
void sfWindowImplWin32::MakeCurrent() const
{
    if (myDeviceContext && myGLContext)
        wglMakeCurrent(myDeviceContext, myGLContext);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::UseVerticalSync
////////////////////////////////////////////////////////////
void sfWindowImplWin32::UseVerticalSync(bool Enabled)
{
    PFNWGLSWAPINTERVALEXTPROC wglSwapIntervalEXT = reinterpret_cast<PFNWGLSWAPINTERVALEXTPROC>(wglGetProcAddress("wglSwapIntervalEXT"));
    if (wglSwapIntervalEXT)
        wglSwapIntervalEXT(Enabled ? 1 : 0);
}


////////////////////////////////////////////////////////////
/// /see sfWindowImpl::ShowMouseCursor
////////////////////////////////////////////////////////////
void sfWindowImplWin32::ShowMouseCursor(bool Show)
{
    ShowCursor(Show);
}


////////////////////////////////////////////////////////////
/// Register the window class
////////////////////////////////////////////////////////////
void sfWindowImplWin32::RegisterWindowClass()
{
    WNDCLASSW WindowClass;
    WindowClass.style         = 0;
    WindowClass.lpfnWndProc   = &sfWindowImplWin32::GlobalOnEvent;
    WindowClass.cbClsExtra    = 0;
    WindowClass.cbWndExtra    = 0;
    WindowClass.hInstance     = GetModuleHandle(NULL);
    WindowClass.hIcon         = NULL;
    WindowClass.hCursor       = 0;
    WindowClass.hbrBackground = 0;
    WindowClass.lpszMenuName  = NULL;
    WindowClass.lpszClassName = ourClassName;

    RegisterClass(&WindowClass);
}


////////////////////////////////////////////////////////////
/// Create the OpenGL rendering context
////////////////////////////////////////////////////////////
void sfWindowImplWin32::CreateContext(sfVideoMode Mode, bool Fullscreen)
{
    // Fullscreen mode
    if (Fullscreen)
    {
        DEVMODE DevMode;
        DevMode.dmSize       = sizeof(DEVMODE);
        DevMode.dmPelsWidth  = Mode.Width;
        DevMode.dmPelsHeight = Mode.Height;
        DevMode.dmBitsPerPel = Mode.BitsPerPixel;
        DevMode.dmFields     = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;

        // Apply fullscreen mode
        if (ChangeDisplaySettings(&DevMode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
        {
            std::cerr << "Failed to change display mode for fullscreen -- cannot create OpenGL context" << std::endl;
            return;
        }

        // Change window style (no border, no titlebar, ...)
        SetWindowLong(myHandle, GWL_STYLE,   WS_POPUP);
        SetWindowLong(myHandle, GWL_EXSTYLE, WS_EX_APPWINDOW);

        // And resize it so that it fits the entire screen
        SetWindowPos(myHandle, HWND_TOP, 0, 0, Mode.Width, Mode.Height, SWP_FRAMECHANGED);
        ShowWindow(myHandle, SW_SHOW);

        // Set "this" as the current fullscreen window
        ourFullscreenWindow = this;
    }

    // SetPixelFormat can fail (really ?) if window style doesn't contain these flags
    long Style = GetWindowLong(myHandle, GWL_STYLE);
    SetWindowLong(myHandle, GWL_STYLE, Style | WS_CLIPCHILDREN | WS_CLIPSIBLINGS);

    // Get the device context attached to the window
    myDeviceContext = GetDC(myHandle);
    if (myDeviceContext == NULL)
    {
        std::cerr << "Failed to get device context of window -- cannot create OpenGL context" << std::endl;
        return;
    }

    // Setup a pixel format descriptor from the rendering settings
    PIXELFORMATDESCRIPTOR PixelDescriptor;
    ZeroMemory(&PixelDescriptor, sizeof(PIXELFORMATDESCRIPTOR));
    PixelDescriptor.nSize        = sizeof(PIXELFORMATDESCRIPTOR);
    PixelDescriptor.nVersion     = 1;
    PixelDescriptor.iLayerType   = PFD_MAIN_PLANE;
    PixelDescriptor.dwFlags      = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    PixelDescriptor.iPixelType   = PFD_TYPE_RGBA;
    PixelDescriptor.cColorBits   = static_cast<BYTE>(Mode.BitsPerPixel);
    PixelDescriptor.cDepthBits   = 32; // It's only a hint
    PixelDescriptor.cStencilBits = 8;  // It's only a hint

    // Get the pixel format that best matches our requirements
    int Index = ChoosePixelFormat(myDeviceContext, &PixelDescriptor);
    if (Index == 0)
    {
        std::cerr << "Failed to find a suitable pixel format for device context -- cannot create OpenGL context" << std::endl;
        return;
    }

    // Extract the depth and stencil bits from the chosen format
    DescribePixelFormat(myDeviceContext, Index, sizeof(PIXELFORMATDESCRIPTOR), &PixelDescriptor);
    myDepthBits   = PixelDescriptor.cDepthBits;
    myStencilBits = PixelDescriptor.cStencilBits;

    // Set the chosen pixel format
    if (!SetPixelFormat(myDeviceContext, Index, &PixelDescriptor))
    {
        std::cerr << "Failed to set pixel format for device context -- cannot create OpenGL context" << std::endl;
        return;
    }

    // Create the OpenGL context from the device context
    myGLContext = wglCreateContext(myDeviceContext);
    if (myGLContext == NULL)
    {
        std::cerr << "Failed to create an OpenGL context for this window" << std::endl;
        return;
    }

    // Share display lists with other contexts
    if (wglGetCurrentContext())
        wglShareLists(wglGetCurrentContext(), myGLContext);

    // Set our context as the current OpenGL context for rendering
    SetCurrent();
}


////////////////////////////////////////////////////////////
/// Free all the graphical resources attached to the window
////////////////////////////////////////////////////////////
void sfWindowImplWin32::Cleanup()
{
    // Restore the previous video mode (in case we were running in fullscreen)
    if (ourFullscreenWindow == this)
    {
        ChangeDisplaySettings(NULL, 0);
        ourFullscreenWindow = NULL;
    }

    // Unhide the mouse cursor (in case it was hidden)
    ShowMouseCursor(true);

    // Destroy the OpenGL context
    if (myGLContext)
    {
        // If this is not the dummy window, we must set it as the valid rendering context to avoid a crash with next OpenGL command
        if (this != ourDummyWindow)
        {
            if (ourDummyWindow)
                ourDummyWindow->SetCurrent();
        }
        else
        {
            ourDummyWindow = NULL;
            wglMakeCurrent(NULL, NULL);
        }

        // Destroy the context
        wglDeleteContext(myGLContext);
        myGLContext = NULL;
    }

    // Release the device context
    if (myDeviceContext)
    {
        ReleaseDC(myHandle, myDeviceContext);
        myDeviceContext = NULL;
    }
}


////////////////////////////////////////////////////////////
/// Process a Win32 event
////////////////////////////////////////////////////////////
void sfWindowImplWin32::ProcessEvent(UINT Message, WPARAM WParam, LPARAM LParam)
{
    // Don't process any message until window is created
    if (myHandle == NULL)
        return;

    switch (Message)
    {
        // Destroy event
        case WM_DESTROY :
        {
            // Here we must cleanup resources !
            Cleanup();
            break;
        }

        // Close event
        case WM_CLOSE :
        {
            sfEvent Event;
            Event.Type = sfEvent::Close;
            SendEvent(Event);
            break;
        }

        // Resize event
        case WM_SIZE :
        {
            // Update window size
            RECT Rect;
            GetClientRect(myHandle, &Rect);
            myWidth  = Rect.right - Rect.left;
            myHeight = Rect.bottom - Rect.top;

            sfEvent Event;
            Event.Type        = sfEvent::Resize;
            Event.Size.Width  = myWidth;
            Event.Size.Height = myHeight;
            SendEvent(Event);
            break;
        }

        // Gain focus event
        case WM_SETFOCUS :
        {
            sfEvent Event;
            Event.Type = sfEvent::GainedFocus;
            SendEvent(Event);
            break;
        }

        // Lost focus event
        case WM_KILLFOCUS :
        {
            sfEvent Event;
            Event.Type = sfEvent::LostFocus;
            SendEvent(Event);
            break;
        }

        // Text event
        case WM_CHAR :
        {
            sfEvent Event;
            Event.Type = sfEvent::TextEntered;
            Event.Text.Unicode = static_cast<sfUint16>(WParam);

            SendEvent(Event);
            break;
        }

        // Keydown event
        case WM_KEYDOWN :
        case WM_SYSKEYDOWN :
        {
            sfEvent Event;
            Event.Type        = sfEvent::KeyPressed;
            Event.Key.Code    = VirtualKeyCodeToSF(WParam);
            Event.Key.Alt     = GetAsyncKeyState(VK_RMENU)    || GetAsyncKeyState(VK_LMENU);
            Event.Key.Control = GetAsyncKeyState(VK_RCONTROL) || GetAsyncKeyState(VK_LCONTROL);
            Event.Key.Shift   = GetAsyncKeyState(VK_RSHIFT)   || GetAsyncKeyState(VK_LSHIFT);

            SendEvent(Event);
            break;
        }

        // Keyup event
        case WM_KEYUP :
        case WM_SYSKEYUP :
        {
            sfEvent Event;
            Event.Type        = sfEvent::KeyReleased;
            Event.Key.Code    = VirtualKeyCodeToSF(WParam);
            Event.Key.Alt     = GetAsyncKeyState(VK_RMENU)    || GetAsyncKeyState(VK_LMENU);
            Event.Key.Control = GetAsyncKeyState(VK_RCONTROL) || GetAsyncKeyState(VK_LCONTROL);
            Event.Key.Shift   = GetAsyncKeyState(VK_RSHIFT)   || GetAsyncKeyState(VK_LSHIFT);

            SendEvent(Event);
            break;
        }

        // Mouse left button down event
        case WM_LBUTTONDOWN :
        {
            sfEvent Event;
            Event.Type          = sfEvent::MouseButtonPressed;
            Event.Mouse.Buttons = sfMouse::Left;
            Event.Mouse.X       = LOWORD(LParam);
            Event.Mouse.Y       = HIWORD(LParam);
            SendEvent(Event);
            break;
        }

        // Mouse left button up event
        case WM_LBUTTONUP :
        {
            sfEvent Event;
            Event.Type          = sfEvent::MouseButtonReleased;
            Event.Mouse.Buttons = sfMouse::Left;
            Event.Mouse.X       = LOWORD(LParam);
            Event.Mouse.Y       = HIWORD(LParam);
            SendEvent(Event);
            break;
        }

        // Mouse right button down event
        case WM_RBUTTONDOWN :
        {
            sfEvent Event;
            Event.Type          = sfEvent::MouseButtonPressed;
            Event.Mouse.Buttons = sfMouse::Right;
            Event.Mouse.X       = LOWORD(LParam);
            Event.Mouse.Y       = HIWORD(LParam);
            SendEvent(Event);
            break;
        }

        // Mouse right button up event
        case WM_RBUTTONUP :
        {
            sfEvent Event;
            Event.Type          = sfEvent::MouseButtonReleased;
            Event.Mouse.Buttons = sfMouse::Right;
            Event.Mouse.X       = LOWORD(LParam);
            Event.Mouse.Y       = HIWORD(LParam);
            SendEvent(Event);
            break;
        }

        // Mouse wheel button down event
        case WM_MBUTTONDOWN :
        {
            sfEvent Event;
            Event.Type          = sfEvent::MouseButtonPressed;
            Event.Mouse.Buttons = sfMouse::Middle;
            Event.Mouse.X       = LOWORD(LParam);
            Event.Mouse.Y       = HIWORD(LParam);
            SendEvent(Event);
            break;
        }

        // Mouse wheel button up event
        case WM_MBUTTONUP :
        {
            sfEvent Event;
            Event.Type          = sfEvent::MouseButtonReleased;
            Event.Mouse.Buttons = sfMouse::Middle;
            Event.Mouse.X       = LOWORD(LParam);
            Event.Mouse.Y       = HIWORD(LParam);
            SendEvent(Event);
            break;
        }

        // Mouse move event
        case WM_MOUSEMOVE :
        {
            sfEvent Event;
            Event.Type = sfEvent::MouseMove;
            Event.Mouse.Buttons = 0;
            if (GetAsyncKeyState(VK_LBUTTON)) Event.Mouse.Buttons |= sfMouse::Left;
            if (GetAsyncKeyState(VK_RBUTTON)) Event.Mouse.Buttons |= sfMouse::Right;
            if (GetAsyncKeyState(VK_MBUTTON)) Event.Mouse.Buttons |= sfMouse::Middle;
            Event.Mouse.X = LOWORD(LParam);
            Event.Mouse.Y = HIWORD(LParam);
            SendEvent(Event);
            break;
        }

        // Joystick XY move event
        case MM_JOY1MOVE :
        case MM_JOY2MOVE :
        {
            sfEvent Event;
            Event.Type                = sfEvent::JoystickMove;
            Event.Joystick.JoystickId = Message == MM_JOY1MOVE ? 0 : 1;

            JOYINFO JoyState;
            joyGetPos(Message == MM_JOY1MOVE ? JOYSTICKID1 : JOYSTICKID2, &JoyState);

            JoyCaps Caps = myJoystickCaps[Event.Joystick.JoystickId];
            Event.Joystick.X = (LOWORD(LParam) - Caps.CenterX) * 200 / Caps.RangeX;
            Event.Joystick.Y = (HIWORD(LParam) - Caps.CenterY) * 200 / Caps.RangeY;
            Event.Joystick.Z = (JoyState.wZpos - Caps.CenterZ) * 200 / Caps.RangeZ;

            SendEvent(Event);
            break;
        }

        // Joystick Z move event
        case MM_JOY1ZMOVE :
        case MM_JOY2ZMOVE :
        {
            sfEvent Event;
            Event.Type                = sfEvent::JoystickMove;
            Event.Joystick.JoystickId = Message == MM_JOY1ZMOVE ? 0 : 1;

            JOYINFO JoyState;
            joyGetPos(Message == MM_JOY1MOVE ? JOYSTICKID1 : JOYSTICKID2, &JoyState);

            JoyCaps Caps = myJoystickCaps[Event.Joystick.JoystickId];
            Event.Joystick.X = (JoyState.wXpos - Caps.CenterX) * 200 / Caps.RangeX;
            Event.Joystick.Y = (JoyState.wYpos - Caps.CenterY) * 200 / Caps.RangeY;
            Event.Joystick.Z = (LOWORD(LParam) - Caps.CenterZ) * 200 / Caps.RangeZ;

            SendEvent(Event);
            break;
        }

        // Joystick button down events
        case MM_JOY1BUTTONDOWN :
        case MM_JOY2BUTTONDOWN :
        {
            sfEvent Event;
            Event.Type                = sfEvent::JoystickButtonPressed;
            Event.Joystick.JoystickId = Message == MM_JOY1BUTTONDOWN ? 0 : 1;

            JOYINFO JoyState;
            joyGetPos(Message == MM_JOY1MOVE ? JOYSTICKID1 : JOYSTICKID2, &JoyState);

            JoyCaps Caps = myJoystickCaps[Event.Joystick.JoystickId];
            Event.Joystick.X = (LOWORD(LParam) - Caps.CenterX) * 200 / Caps.RangeX;
            Event.Joystick.Y = (HIWORD(LParam) - Caps.CenterY) * 200 / Caps.RangeY;
            Event.Joystick.Z = (JoyState.wZpos - Caps.CenterZ) * 200 / Caps.RangeZ;

            if      ((WParam & JOY_BUTTON1) && (WParam & JOY_BUTTON1CHG)) Event.Joystick.Button = 0;
            else if ((WParam & JOY_BUTTON2) && (WParam & JOY_BUTTON2CHG)) Event.Joystick.Button = 1;
            else if ((WParam & JOY_BUTTON3) && (WParam & JOY_BUTTON3CHG)) Event.Joystick.Button = 2;
            else if ((WParam & JOY_BUTTON4) && (WParam & JOY_BUTTON4CHG)) Event.Joystick.Button = 3;

            SendEvent(Event);
            break;
        }

        // Joystick button up events
        case MM_JOY1BUTTONUP :
        case MM_JOY2BUTTONUP :
        {
            sfEvent Event;
            Event.Type                = sfEvent::JoystickButtonReleased;
            Event.Joystick.JoystickId = Message == MM_JOY1BUTTONUP ? 0 : 1;

            JOYINFO JoyState;
            joyGetPos(Message == MM_JOY1MOVE ? JOYSTICKID1 : JOYSTICKID2, &JoyState);

            JoyCaps Caps = myJoystickCaps[Event.Joystick.JoystickId];
            Event.Joystick.X = (LOWORD(LParam) - Caps.CenterX) * 200 / Caps.RangeX;
            Event.Joystick.Y = (HIWORD(LParam) - Caps.CenterY) * 200 / Caps.RangeY;
            Event.Joystick.Z = (JoyState.wZpos - Caps.CenterZ) * 200 / Caps.RangeZ;

            if      (!(WParam & JOY_BUTTON1) && (WParam & JOY_BUTTON1CHG)) Event.Joystick.Button = 0;
            else if (!(WParam & JOY_BUTTON2) && (WParam & JOY_BUTTON2CHG)) Event.Joystick.Button = 1;
            else if (!(WParam & JOY_BUTTON3) && (WParam & JOY_BUTTON3CHG)) Event.Joystick.Button = 2;
            else if (!(WParam & JOY_BUTTON4) && (WParam & JOY_BUTTON4CHG)) Event.Joystick.Button = 3;

            SendEvent(Event);
            break;
        }
    }
}


////////////////////////////////////////////////////////////
/// Setup the joysticks
////////////////////////////////////////////////////////////
void sfWindowImplWin32::SetupJoysticks()
{
    // Get the number of connected joysticks
    MMRESULT Error;
    JOYINFOEX JoyInfo;
    unsigned int Count = 0;
    while ((Count < 2) && (Error = joyGetPosEx(JOYSTICKID1 + Count, &JoyInfo)) != JOYERR_PARMS)
    {
        // Check if the current joystick is connected
        if (Error == JOYERR_NOERROR)
        {
            // Start capturing joystick
            joySetCapture(myHandle, JOYSTICKID1 + Count, 10, TRUE);

            // Store its caps
            JOYCAPS WinCaps;
            joyGetDevCaps(JOYSTICKID1 + Count, &WinCaps, sizeof(WinCaps));
            myJoystickCaps[Count].CenterX = (WinCaps.wXmax + WinCaps.wXmin) / 2;
            myJoystickCaps[Count].CenterY = (WinCaps.wYmax + WinCaps.wYmin) / 2;
            myJoystickCaps[Count].CenterZ = (WinCaps.wZmax + WinCaps.wZmin) / 2;
            myJoystickCaps[Count].RangeX  = WinCaps.wXmax - WinCaps.wXmin;
            myJoystickCaps[Count].RangeY  = WinCaps.wYmax - WinCaps.wYmin;
            myJoystickCaps[Count].RangeZ  = WinCaps.wZmax - WinCaps.wZmin;
        }

        // Go to the next joystick
        ++Count;
    }
}


////////////////////////////////////////////////////////////
/// Convert a Win32 virtual key code to a SFML key code
////////////////////////////////////////////////////////////
sfKey::Code sfWindowImplWin32::VirtualKeyCodeToSF(WPARAM VirtualKey)
{
    switch (VirtualKey)
    {
        case VK_ESCAPE :   return sfKey::Escape;
        case VK_SPACE :    return sfKey::Space;
        case VK_RETURN :   return sfKey::Return;
        case VK_BACK :     return sfKey::Back;
        case VK_TAB :      return sfKey::Tab;
        case VK_PRIOR :    return sfKey::PageUp;
        case VK_NEXT :     return sfKey::PageDown;
        case VK_END :      return sfKey::End;
        case VK_HOME :     return sfKey::Home;
        case VK_INSERT :   return sfKey::Insert;
        case VK_DELETE :   return sfKey::Delete;
        case VK_ADD :      return sfKey::Add;
        case VK_SUBTRACT : return sfKey::Subtract;
        case VK_MULTIPLY : return sfKey::Multiply;
        case VK_DIVIDE :   return sfKey::Divide;
        case VK_F1 :       return sfKey::F1;
        case VK_F2 :       return sfKey::F2;
        case VK_F3 :       return sfKey::F3;
        case VK_F4 :       return sfKey::F4;
        case VK_F5 :       return sfKey::F5;
        case VK_F6 :       return sfKey::F6;
        case VK_F7 :       return sfKey::F7;
        case VK_F8 :       return sfKey::F8;
        case VK_F9 :       return sfKey::F9;
        case VK_F10 :      return sfKey::F10;
        case VK_F11 :      return sfKey::F11;
        case VK_F12 :      return sfKey::F12;
        case VK_F13 :      return sfKey::F13;
        case VK_F14 :      return sfKey::F14;
        case VK_F15 :      return sfKey::F15;
        case VK_LEFT :     return sfKey::Left;
        case VK_RIGHT :    return sfKey::Right;
        case VK_UP :       return sfKey::Up;
        case VK_DOWN :     return sfKey::Down;
        case VK_NUMPAD0 :  return sfKey::Numpad0;
        case VK_NUMPAD1 :  return sfKey::Numpad1;
        case VK_NUMPAD2 :  return sfKey::Numpad2;
        case VK_NUMPAD3 :  return sfKey::Numpad3;
        case VK_NUMPAD4 :  return sfKey::Numpad4;
        case VK_NUMPAD5 :  return sfKey::Numpad5;
        case VK_NUMPAD6 :  return sfKey::Numpad6;
        case VK_NUMPAD7 :  return sfKey::Numpad7;
        case VK_NUMPAD8 :  return sfKey::Numpad8;
        case VK_NUMPAD9 :  return sfKey::Numpad9;
        case 'A' :         return sfKey::A;
        case 'Z' :         return sfKey::Z;
        case 'E' :         return sfKey::E;
        case 'R' :         return sfKey::R;
        case 'T' :         return sfKey::T;
        case 'Y' :         return sfKey::Y;
        case 'U' :         return sfKey::U;
        case 'I' :         return sfKey::I;
        case 'O' :         return sfKey::O;
        case 'P' :         return sfKey::P;
        case 'Q' :         return sfKey::Q;
        case 'S' :         return sfKey::S;
        case 'D' :         return sfKey::D;
        case 'F' :         return sfKey::F;
        case 'G' :         return sfKey::G;
        case 'H' :         return sfKey::H;
        case 'J' :         return sfKey::J;
        case 'K' :         return sfKey::K;
        case 'L' :         return sfKey::L;
        case 'M' :         return sfKey::M;
        case 'W' :         return sfKey::W;
        case 'X' :         return sfKey::X;
        case 'C' :         return sfKey::C;
        case 'V' :         return sfKey::V;
        case 'B' :         return sfKey::B;
        case 'N' :         return sfKey::N;
        case '0' :         return sfKey::Num0;
        case '1' :         return sfKey::Num1;
        case '2' :         return sfKey::Num2;
        case '3' :         return sfKey::Num3;
        case '4' :         return sfKey::Num4;
        case '5' :         return sfKey::Num5;
        case '6' :         return sfKey::Num6;
        case '7' :         return sfKey::Num7;
        case '8' :         return sfKey::Num8;
        case '9' :         return sfKey::Num9;
    }

    return sfKey::Code(0);
}


////////////////////////////////////////////////////////////
/// Function called whenever one of our windows receives a message
////////////////////////////////////////////////////////////
LRESULT CALLBACK sfWindowImplWin32::GlobalOnEvent(HWND Handle, UINT Message, WPARAM WParam, LPARAM LParam)
{
    // Associate handle and Window instance when the creation message is received
    if (Message == WM_CREATE)
    {
        // Get sfWindowImplWin32 instance (it was passed as the last argument of CreateWindow)
        long This = reinterpret_cast<long>(reinterpret_cast<CREATESTRUCT*>(LParam)->lpCreateParams);

        // Set as the "user data" parameter of the window
        SetWindowLongPtr(Handle, GWLP_USERDATA, This);
    }

    // Get the WindowImpl instance corresponding to the window handle
    sfWindowImplWin32* Window = reinterpret_cast<sfWindowImplWin32*>(GetWindowLongPtr(Handle, GWLP_USERDATA));

    // Forward the event to the appropriate function
    if (Window)
    {
        Window->ProcessEvent(Message, WParam, LParam);

        if (Window->myCallback)
            return CallWindowProc(reinterpret_cast<WNDPROC>(Window->myCallback), Handle, Message, WParam, LParam);
    }

    return DefWindowProc(Handle, Message, WParam, LParam);
}

} // namespace sf_private
