////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Window/Window.hpp>
#include <SFML/Window/WindowImpl.hpp>
#include <iostream>
#include <memory>

namespace
{
    // Create a global dummy window, so that we have a valid OpenGL context at program startup
    //
    // TODO : provide a way to control the dummy creation / destruction, as the order of
    // initialization of globals is completely random across compile units...
    //
    std::auto_ptr<sf_private::sfWindowImpl> DummyWindow(sf_private::sfWindowImpl::New());
}


////////////////////////////////////////////////////////////
/// Default constructor
///
////////////////////////////////////////////////////////////
sfWindow::sfWindow() :
myWindow       (NULL),
myLastFrameTime(0.f)
{

}


////////////////////////////////////////////////////////////
/// Construct a new window
////////////////////////////////////////////////////////////
sfWindow::sfWindow(sfVideoMode Mode, const std::string& Title, bool Fullscreen) :
myWindow       (NULL),
myLastFrameTime(0.f)
{
    Create(Mode, Title, Fullscreen);
}


////////////////////////////////////////////////////////////
/// Construct the window from an existing control
////////////////////////////////////////////////////////////
sfWindow::sfWindow(sfWindowHandle Handle) :
myWindow       (NULL),
myLastFrameTime(0.f)
{
    Create(Handle);
}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
sfWindow::~sfWindow()
{
    // Destroy the window implementation
    delete myWindow;
}


////////////////////////////////////////////////////////////
/// Create the window
////////////////////////////////////////////////////////////
void sfWindow::Create(sfVideoMode Mode, const std::string& Title, bool Fullscreen)
{
    // Check validity of video mode
    if (Fullscreen && !Mode.IsValid())
    {
        std::cerr << "Requested video mode is not available, switching to a valid mode" << std::endl;
        Mode = sfVideoMode::GetMode(0);
    }

    Initialize(sf_private::sfWindowImpl::New(Mode, Title, Fullscreen));
}


////////////////////////////////////////////////////////////
/// Create the window from an existing control
////////////////////////////////////////////////////////////
void sfWindow::Create(sfWindowHandle Handle)
{
    Initialize(sf_private::sfWindowImpl::New(Handle));
}


////////////////////////////////////////////////////////////
/// Get the width of the rendering region of the window
////////////////////////////////////////////////////////////
unsigned int sfWindow::GetWidth() const
{
    return myWindow ? myWindow->GetWidth() : 0;
}


////////////////////////////////////////////////////////////
/// Get the height of the rendering region of the window
////////////////////////////////////////////////////////////
unsigned int sfWindow::GetHeight() const
{
    return myWindow ? myWindow->GetHeight() : 0;
}


////////////////////////////////////////////////////////////
/// Get the depth buffer bits
////////////////////////////////////////////////////////////
unsigned int sfWindow::GetDepthBits() const
{
    return myWindow ? myWindow->GetDepthBits() : 0;
}


////////////////////////////////////////////////////////////
/// Get the stencil buffer bits
////////////////////////////////////////////////////////////
unsigned int sfWindow::GetStencilBits() const
{
    return myWindow ? myWindow->GetStencilBits() : 0;
}


////////////////////////////////////////////////////////////
/// Get the event on top of events stack, if any
////////////////////////////////////////////////////////////
bool sfWindow::GetEvent(sfEvent& Event)
{
    // Pop first event of queue, if not empty
    if (!myEvents.empty())
    {
        Event = myEvents.front();
        myEvents.pop();

        return true;
    }

    return false;
}


////////////////////////////////////////////////////////////
/// Enable / disable vertical synchronization
////////////////////////////////////////////////////////////
void sfWindow::UseVerticalSync(bool Enabled)
{
    if (SetCurrent())
        myWindow->UseVerticalSync(Enabled);
}


////////////////////////////////////////////////////////////
/// Show or hide the mouse cursor
////////////////////////////////////////////////////////////
void sfWindow::ShowMouseCursor(bool Show)
{
    if (SetCurrent())
        myWindow->ShowMouseCursor(Show);
}


////////////////////////////////////////////////////////////
/// Set the window as the current target for rendering
////////////////////////////////////////////////////////////
bool sfWindow::SetCurrent() const
{
    if (myWindow)
    {
        myWindow->SetCurrent();
        return true;
    }

    return false;
}


////////////////////////////////////////////////////////////
/// Display the window on screen
////////////////////////////////////////////////////////////
void sfWindow::Display()
{
    // Measure the time elapsed since last frame
    myLastFrameTime = myClock.GetElapsedTime();
    myClock.Reset();

    if (SetCurrent())
    {
        // Display the backbuffer on screen
        myWindow->Display();

        // Let the window implementation process incoming events
        myWindow->ProcessEvents();
    }
}


////////////////////////////////////////////////////////////
/// Get the input manager of the window
////////////////////////////////////////////////////////////
const sfInput& sfWindow::GetInput() const
{
    return myInput;
}


////////////////////////////////////////////////////////////
/// Get time elapsed since last frame
////////////////////////////////////////////////////////////
float sfWindow::GetFrameTime() const
{
    return myLastFrameTime;
}


////////////////////////////////////////////////////////////
/// Receive an event from window
////////////////////////////////////////////////////////////
void sfWindow::OnEvent(const sfEvent& Event)
{
    myEvents.push(Event);
}


////////////////////////////////////////////////////////////
/// Initialize internal window
////////////////////////////////////////////////////////////
void sfWindow::Initialize(sf_private::sfWindowImpl* Window)
{
    // Destroy window if already created
    if (myWindow)
        delete myWindow;

    // Assign new window and listen to its events
    myWindow = Window;
    myWindow->AddListener(this);

    // Attach input to the window
    myWindow->AddListener(&myInput);

    // Disable vertical synchronization by default (to get a consistent behavior)
    UseVerticalSync(false);

    // Reset frame time
    myClock.Reset();
    myLastFrameTime = 0.f;
}
