////////////////////////////////////////////////////////////
//
// SFML - Simple and Fast Multimedia Library
// Copyright (C) 2007 Laurent Gomila (laurent.gom@gmail.com)
//
// This software is provided 'as-is', without any express or implied warranty.
// In no event will the authors be held liable for any damages arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented;
//    you must not claim that you wrote the original software.
//    If you use this software in a product, an acknowledgment
//    in the product documentation would be appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such,
//    and must not be misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source distribution.
//
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// Headers
////////////////////////////////////////////////////////////
#include <SFML/Window/WindowImpl.hpp>
#include <SFML/Window/Event.hpp>
#include <SFML/Window/OpenGLCaps.hpp>
#include <SFML/Window/WindowListener.hpp>
#include <algorithm>

#if defined(SFML_SYSTEM_WINDOWS)

    #include <SFML/Window/Win32/WindowImplWin32.hpp>

#elif defined(SFML_SYSTEM_LINUX) || defined(SFML_SYSTEM_MACOS)

    #include <SFML/Window/Unix/WindowImplUnix.hpp>

#endif


namespace sf_private
{
////////////////////////////////////////////////////////////
// Static member data
////////////////////////////////////////////////////////////
const sfWindowImpl* sfWindowImpl::ourCurrent = NULL;


////////////////////////////////////////////////////////////
/// Create a new window depending on the current OS
////////////////////////////////////////////////////////////
sfWindowImpl* sfWindowImpl::New()
{
    sfWindowImpl* Window = NULL;

    #if defined(SFML_SYSTEM_WINDOWS)

        // Win32 window
        Window = new sfWindowImplWin32();

    #elif defined(SFML_SYSTEM_LINUX) || defined(SFML_SYSTEM_MACOS)

        // Unix window
        Window = new sfWindowImplUnix();

    #endif

    // This is supposed to be the first window created,
    // so once it is done we can initialize the OpenGL caps
    sfOpenGLCaps::Initialize();

    return Window;
}


////////////////////////////////////////////////////////////
/// Create a new window depending on the current OS
////////////////////////////////////////////////////////////
sfWindowImpl* sfWindowImpl::New(sfVideoMode Mode, const std::string& Title, bool Fullscreen)
{
    #if defined(SFML_SYSTEM_WINDOWS)

        // Win32 window
        return new sfWindowImplWin32(Mode, Title, Fullscreen);

    #elif defined(SFML_SYSTEM_LINUX) || defined(SFML_SYSTEM_MACOS)

        // Unix window
        return new sfWindowImplUnix(Mode, Title, Fullscreen);

    #endif
}


////////////////////////////////////////////////////////////
/// Create a new window depending on the current OS
////////////////////////////////////////////////////////////
sfWindowImpl* sfWindowImpl::New(sfWindowHandle Handle)
{
    #if defined(SFML_SYSTEM_WINDOWS)

        // Win32 window
        return new sfWindowImplWin32(Handle);

    #elif defined(SFML_SYSTEM_LINUX) || defined(SFML_SYSTEM_MACOS)

        // Unix window
        return new sfWindowImplUnix(Handle);

    #endif
}


////////////////////////////////////////////////////////////
/// Default constructor
////////////////////////////////////////////////////////////
sfWindowImpl::sfWindowImpl() :
myWidth      (0),
myHeight     (0),
myDepthBits  (0),
myStencilBits(0)
{

}


////////////////////////////////////////////////////////////
/// Destructor
////////////////////////////////////////////////////////////
sfWindowImpl::~sfWindowImpl()
{
    // Nothing to do
}


////////////////////////////////////////////////////////////
/// Add a listener to the window
////////////////////////////////////////////////////////////
void sfWindowImpl::AddListener(sfWindowListener* Listener)
{
    if (Listener)
        myListeners.insert(Listener);
}


////////////////////////////////////////////////////////////
/// Remove a listener from the window
////////////////////////////////////////////////////////////
void sfWindowImpl::RemoveListener(sfWindowListener* Listener)
{
    myListeners.erase(Listener);
}


////////////////////////////////////////////////////////////
/// Get the client width of the window
////////////////////////////////////////////////////////////
unsigned int sfWindowImpl::GetWidth() const
{
    return myWidth;
}


////////////////////////////////////////////////////////////
/// Get the client height of the window
////////////////////////////////////////////////////////////
unsigned int sfWindowImpl::GetHeight() const
{
    return myHeight;
}


////////////////////////////////////////////////////////////
/// Get the depth buffer bits
////////////////////////////////////////////////////////////
unsigned int sfWindowImpl::GetDepthBits() const
{
    return myDepthBits;
}


////////////////////////////////////////////////////////////
/// Get the stencil buffer bits
////////////////////////////////////////////////////////////
unsigned int sfWindowImpl::GetStencilBits() const
{
    return myStencilBits;
}


////////////////////////////////////////////////////////////
/// Set the window as the current target for rendering
////////////////////////////////////////////////////////////
void sfWindowImpl::SetCurrent() const
{
    if (ourCurrent != this)
    {
        MakeCurrent();
        ourCurrent = this;
    }
}


////////////////////////////////////////////////////////////
/// Send an event to listeners
////////////////////////////////////////////////////////////
void sfWindowImpl::SendEvent(const sfEvent& Event)
{
    for (std::set<sfWindowListener*>::iterator i = myListeners.begin(); i != myListeners.end(); ++i)
    {
        (*i)->OnEvent(Event);
    }
}

} // namespace sf_private
